from typing import Optional

from pydantic import Field

from datahub.configuration.common import (
    AllowDenyPattern,
    ConfigModel,
    TransparentSecretStr,
)
from datahub.configuration.source_common import (
    EnvConfigMixin,
    PlatformInstanceConfigMixin,
)
from datahub.ingestion.source.ge_profiling_config import GEProfilingBaseConfig
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
)
from datahub.ingestion.source_config.operation_config import is_profiling_enabled

# - Referencing https://docs.datastax.com/en/cql-oss/3.x/cql/cql_using/useQuerySystem.html#Table3.ColumnsinSystem_SchemaTables-Cassandra3.0 - #
# this keyspace contains details about the cassandra cluster's keyspaces, tables, and columns
SYSTEM_SCHEMA_KEYSPACE_NAME = "system_schema"

# Reference:
# https://docs.datastax.com/en/astra-db-serverless/databases/python-driver.html
# https://docs.datastax.com/en/astra-db-serverless/databases/python-driver.html#production-configuration


class CassandraCloudConfig(ConfigModel):
    """
    Configuration for connecting to DataStax Astra DB in the cloud.
    """

    token: TransparentSecretStr = Field(
        description="The Astra DB application token used for authentication.",
    )

    secure_connect_bundle: str = Field(
        description="File path to the Secure Connect Bundle (.zip) used for a secure connection to DataStax Astra DB.",
    )

    connect_timeout: int = Field(
        default=600,
        description="Timeout in seconds for establishing new connections to Cassandra.",
    )

    request_timeout: int = Field(
        default=600, description="Timeout in seconds for individual Cassandra requests."
    )


class CassandraSourceConfig(
    PlatformInstanceConfigMixin, StatefulIngestionConfigBase, EnvConfigMixin
):
    """
    Configuration for connecting to a Cassandra or DataStax Astra DB source.
    """

    contact_point: str = Field(
        default="localhost",
        description="Domain or IP address of the Cassandra instance (excluding port).",
    )

    port: int = Field(
        default=9042, description="Port number to connect to the Cassandra instance."
    )

    username: Optional[str] = Field(
        default=None,
        description=f"Username credential with read access to the {SYSTEM_SCHEMA_KEYSPACE_NAME} keyspace.",
    )

    password: Optional[TransparentSecretStr] = Field(
        default=None,
        description="Password credential associated with the specified username.",
    )

    cloud_config: Optional[CassandraCloudConfig] = Field(
        default=None,
        description="Configuration for cloud-based Cassandra, such as DataStax Astra DB.",
    )

    ssl_ca_certs: Optional[str] = Field(
        default=None,
        description="Path to the CA certificate file for SSL connections.",
    )

    ssl_certfile: Optional[str] = Field(
        default=None,
        description="Path to the SSL certificate file for SSL connections.",
    )

    ssl_keyfile: Optional[str] = Field(
        default=None,
        description="Path to the SSL key file for SSL connections.",
    )

    ssl_version: Optional[str] = Field(
        default="TLS_CLIENT",
        description="SSL protocol version to use for connections. Options: TLS_CLIENT, TLSv1, TLSv1_1, TLSv1_2, TLSv1_3. Defaults to TLS_CLIENT.",
    )

    keyspace_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns to filter keyspaces for ingestion.",
    )

    table_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns to filter keyspaces.tables for ingestion.",
    )

    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = Field(
        default=None,
        description="Configuration for stateful ingestion and stale metadata removal.",
    )

    # Profiling
    profile_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for tables to profile",
    )

    profiling: GEProfilingBaseConfig = Field(
        default=GEProfilingBaseConfig(),
        description="Configuration for profiling",
    )

    def is_profiling_enabled(self) -> bool:
        return self.profiling.enabled and is_profiling_enabled(
            self.profiling.operation_config
        )



import warnings
from typing import TYPE_CHECKING, Any, Optional, Type

from pydantic import model_validator

from datahub.configuration.common import ConfigurationWarning
from datahub.utilities.global_warning_util import add_global_warning

if TYPE_CHECKING:
    from pydantic.deprecated.class_validators import V1RootValidator

_unset = object()


def pydantic_field_deprecated(
    field: str,
    warn_if_value_is_not: Any = _unset,
    message: Optional[str] = None,
) -> "V1RootValidator":
    if message:
        output = message
    else:
        output = f"{field} is deprecated and will be removed in a future release. Please remove it from your config."

    def _validate_deprecated(cls: Type, values: dict) -> dict:
        if field in values and (
            warn_if_value_is_not is _unset or values[field] != warn_if_value_is_not
        ):
            add_global_warning(output)
            warnings.warn(output, ConfigurationWarning, stacklevel=2)
        return values

    # Hack: Pydantic maintains unique list of validators by referring its __name__.
    # https://github.com/pydantic/pydantic/blob/v1.10.9/pydantic/main.py#L264
    # This hack ensures that multiple field deprecated do not overwrite each other.
    _validate_deprecated.__name__ = f"{_validate_deprecated.__name__}_{field}"
    return model_validator(mode="before")(_validate_deprecated)



import functools
from typing import Any, Optional, Tuple

import sqlglot
from pydantic import BaseModel


class _ParserBaseModel(
    BaseModel,
    arbitrary_types_allowed=True,
):
    def json(self, *args: Any, **kwargs: Any) -> str:
        return super().model_dump_json(*args, **kwargs)  # type: ignore


@functools.total_ordering
class _FrozenModel(_ParserBaseModel, frozen=True):
    def __lt__(self, other: "_FrozenModel") -> bool:
        for field in self.__class__.model_fields:
            self_v = getattr(self, field)
            other_v = getattr(other, field)

            # Handle None values by pushing them to the end of the ordering.
            if self_v is None and other_v is not None:
                return False
            elif self_v is not None and other_v is None:
                return True
            elif self_v != other_v:
                return self_v < other_v

        return False


class _TableName(_FrozenModel):
    # TODO: Move this into the schema_resolver.py file.

    database: Optional[str] = None
    db_schema: Optional[str] = None
    table: str
    parts: Optional[Tuple[str, ...]] = None

    @property
    def identity(
        self,
    ) -> Tuple[Optional[str], Optional[str], str, Optional[Tuple[str, ...]]]:
        """
        Table identity for hashing and equality.

        Includes parts only when >3 parts to distinguish multi-part tables
        while maintaining backward compatibility for standard 3-part names.
        """
        if self.parts and len(self.parts) > 3:
            return (self.database, self.db_schema, self.table, self.parts)
        return (self.database, self.db_schema, self.table, None)

    def __hash__(self) -> int:
        return hash(self.identity)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _TableName):
            return False
        return self.identity == other.identity

    def as_sqlglot_table(self) -> sqlglot.exp.Table:
        return sqlglot.exp.Table(
            catalog=(
                sqlglot.exp.Identifier(this=self.database) if self.database else None
            ),
            db=sqlglot.exp.Identifier(this=self.db_schema) if self.db_schema else None,
            this=sqlglot.exp.Identifier(this=self.table),
        )

    def qualified(
        self,
        dialect: sqlglot.Dialect,
        default_db: Optional[str] = None,
        default_schema: Optional[str] = None,
    ) -> "_TableName":
        database = self.database or default_db
        db_schema = self.db_schema or default_schema

        return _TableName(
            database=database,
            db_schema=db_schema,
            table=self.table,
            parts=self.parts,
        )

    @classmethod
    def from_sqlglot_table(
        cls,
        table: sqlglot.exp.Table,
        default_db: Optional[str] = None,
        default_schema: Optional[str] = None,
    ) -> "_TableName":
        # Handle Snowflake semantic views: SEMANTIC_VIEW(table_name ...)
        # In this case, table.this is a SemanticView expression, and we need to
        # extract the actual table from within it.
        if isinstance(table.this, sqlglot.exp.SemanticView):
            # The SemanticView.this contains the actual table reference
            inner_table = table.this.this
            if isinstance(inner_table, sqlglot.exp.Table):
                # Recursively extract from the inner table
                return cls.from_sqlglot_table(inner_table, default_db, default_schema)
            elif isinstance(inner_table, sqlglot.exp.Identifier):
                # Simple table name
                return cls(
                    database=table.catalog or default_db,
                    db_schema=table.db or default_schema,
                    table=inner_table.name,
                    parts=None,
                )

        if isinstance(table.this, sqlglot.exp.Dot):
            # Multi-part tables (>3 parts) have extra parts in a Dot expression
            parts = []
            exp = table.this
            while isinstance(exp, sqlglot.exp.Dot):
                parts.append(exp.this.name)
                exp = exp.expression
            parts.append(exp.name)
            table_name = ".".join(parts)
        else:
            table_name = table.this.name

        parts_tuple = tuple(p.name for p in table.parts) if table.parts else None

        return cls(
            database=table.catalog or default_db,
            db_schema=table.db or default_schema,
            table=table_name,
            parts=parts_tuple,
        )



import logging
import os
from typing import Any, Dict, List, Union

from pydantic import BaseModel

from datahub.secret.secret_store import SecretStore

logger = logging.getLogger(__name__)


class FileSecretStoreConfig(BaseModel):
    basedir: str = "/mnt/secrets"
    max_length: int = 1024768


# Simple SecretStore implementation that fetches Secret values from the local files.
class FileSecretStore(SecretStore):
    def __init__(self, config: FileSecretStoreConfig):
        self.config = config

    def get_secret_values(self, secret_names: List[str]) -> Dict[str, Union[str, None]]:
        values = {}
        for secret_name in secret_names:
            values[secret_name] = self.get_secret_value(secret_name)
        return values

    def get_secret_value(self, secret_name: str) -> Union[str, None]:
        secret_path = os.path.join(self.config.basedir, secret_name)
        if os.path.exists(secret_path):
            with open(secret_path, "r") as f:
                secret_value = f.read(self.config.max_length + 1)
                if len(secret_value) > self.config.max_length:
                    logger.warning(
                        f"Secret {secret_name} is longer than {self.config.max_length} and will be truncated."
                    )
                return secret_value[: self.config.max_length].rstrip()
        return None

    def get_id(self) -> str:
        return "file"

    def close(self) -> None:
        return

    @classmethod
    def create(cls, config: Any) -> "FileSecretStore":
        config = FileSecretStoreConfig.model_validate(config)
        return cls(config)



from __future__ import annotations

from datetime import datetime
from typing import Dict, List, Optional, Sequence, Type, Union

from deprecated.sphinx import deprecated
from typing_extensions import Self

import datahub.metadata.schema_classes as models
from datahub.emitter.enum_helpers import get_enum_options
from datahub.metadata.urns import ChartUrn, DatasetUrn, Urn
from datahub.sdk._shared import (
    ActorUrnOrStr,
    ChangeAuditStampsMixin,
    DataPlatformInstanceUrnOrStr,
    DataPlatformUrnOrStr,
    DatasetUrnOrStr,
    DomainInputType,
    HasContainer,
    HasDomain,
    HasInstitutionalMemory,
    HasOwnership,
    HasPlatformInstance,
    HasSubtype,
    HasTags,
    HasTerms,
    LinksInputType,
    OwnersInputType,
    ParentContainerInputType,
    TagsInputType,
    TermsInputType,
)
from datahub.sdk.dataset import Dataset
from datahub.sdk.entity import Entity, ExtraAspectsType
from datahub.utilities.sentinels import Unset, unset


class Chart(
    ChangeAuditStampsMixin,
    HasPlatformInstance,
    HasSubtype,
    HasOwnership,
    HasContainer,
    HasInstitutionalMemory,
    HasTags,
    HasTerms,
    HasDomain,
    Entity,
):
    """Represents a chart in DataHub."""

    __slots__ = ()

    @classmethod
    def get_urn_type(cls) -> Type[ChartUrn]:
        """Get the URN type for charts.
        Returns:
            The ChartUrn class.
        """
        return ChartUrn

    def __init__(
        self,
        *,
        # Identity.
        name: str,
        platform: DataPlatformUrnOrStr,
        display_name: Optional[str] = None,
        platform_instance: Optional[DataPlatformInstanceUrnOrStr] = None,
        # Chart properties.
        description: Optional[str] = "",
        external_url: Optional[str] = None,
        chart_url: Optional[str] = None,
        custom_properties: Optional[Dict[str, str]] = None,
        last_modified: Optional[datetime] = None,
        last_modified_by: Optional[ActorUrnOrStr] = None,
        created_at: Optional[datetime] = None,
        created_by: Optional[ActorUrnOrStr] = None,
        deleted_on: Optional[datetime] = None,
        deleted_by: Optional[ActorUrnOrStr] = None,
        last_refreshed: Optional[datetime] = None,
        chart_type: Optional[Union[str, models.ChartTypeClass]] = None,
        access: Optional[str] = None,
        input_datasets: Optional[Sequence[Union[DatasetUrnOrStr, Dataset]]] = None,
        # Standard aspects.
        parent_container: ParentContainerInputType | Unset = unset,
        subtype: Optional[str] = None,
        owners: Optional[OwnersInputType] = None,
        links: Optional[LinksInputType] = None,
        tags: Optional[TagsInputType] = None,
        terms: Optional[TermsInputType] = None,
        domain: Optional[DomainInputType] = None,
        extra_aspects: ExtraAspectsType = None,
    ):
        """Initialize a new Chart instance."""
        urn = ChartUrn.create_from_ids(
            platform=str(platform),
            name=name,
            platform_instance=str(platform_instance) if platform_instance else None,
        )
        super().__init__(urn)
        self._set_extra_aspects(extra_aspects)

        self._set_platform_instance(platform, platform_instance)
        self._ensure_chart_props(display_name=display_name)
        self._init_chart_properties(
            description,
            display_name,
            external_url,
            chart_url,
            custom_properties,
            last_modified,
            last_modified_by,
            created_at,
            created_by,
            last_refreshed,
            deleted_on,
            deleted_by,
            chart_type,
            access,
            input_datasets,
        )
        self._init_standard_aspects(
            parent_container, subtype, owners, links, tags, terms, domain
        )

    @classmethod
    def _new_from_graph(cls, urn: Urn, current_aspects: models.AspectBag) -> Self:
        assert isinstance(urn, ChartUrn)
        entity = cls(
            platform=urn.dashboard_tool,
            name=urn.chart_id,
        )
        return entity._init_from_graph(current_aspects)

    def _init_chart_properties(
        self,
        description: Optional[str],
        display_name: Optional[str],
        external_url: Optional[str],
        chart_url: Optional[str],
        custom_properties: Optional[Dict[str, str]],
        last_modified: Optional[datetime],
        last_modified_by: Optional[ActorUrnOrStr],
        created_at: Optional[datetime],
        created_by: Optional[ActorUrnOrStr],
        last_refreshed: Optional[datetime],
        deleted_on: Optional[datetime],
        deleted_by: Optional[ActorUrnOrStr],
        chart_type: Optional[Union[str, models.ChartTypeClass]],
        access: Optional[str],
        input_datasets: Optional[Sequence[Union[DatasetUrnOrStr, Dataset]]],
    ) -> None:
        """Initialize chart-specific properties."""
        if description is not None:
            self.set_description(description)
        if display_name is not None:
            self.set_display_name(display_name)
        if external_url is not None:
            self.set_external_url(external_url)
        if chart_url is not None:
            self.set_chart_url(chart_url)
        if custom_properties is not None:
            self.set_custom_properties(custom_properties)
        if last_modified is not None:
            self.set_last_modified(last_modified)
        if last_modified_by is not None:
            self.set_last_modified_by(last_modified_by)
        if created_at is not None:
            self.set_created_at(created_at)
        if created_by is not None:
            self.set_created_by(created_by)
        if deleted_on is not None:
            self.set_deleted_on(deleted_on)
        if deleted_by is not None:
            self.set_deleted_by(deleted_by)
        if last_refreshed is not None:
            self.set_last_refreshed(last_refreshed)
        if chart_type is not None:
            self.set_chart_type(chart_type)
        if access is not None:
            self.set_access(access)
        if input_datasets is not None:
            self.set_input_datasets(input_datasets)

    def _init_standard_aspects(
        self,
        parent_container: ParentContainerInputType | Unset,
        subtype: Optional[str],
        owners: Optional[OwnersInputType],
        links: Optional[LinksInputType],
        tags: Optional[TagsInputType],
        terms: Optional[TermsInputType],
        domain: Optional[DomainInputType],
    ) -> None:
        """Initialize standard aspects."""
        if parent_container is not unset:
            self._set_container(parent_container)
        if subtype is not None:
            self.set_subtype(subtype)
        if owners is not None:
            self.set_owners(owners)
        if links is not None:
            self.set_links(links)
        if tags is not None:
            self.set_tags(tags)
        if terms is not None:
            self.set_terms(terms)
        if domain is not None:
            self.set_domain(domain)

    @property
    def urn(self) -> ChartUrn:
        assert isinstance(self._urn, ChartUrn)
        return self._urn

    def _ensure_chart_props(
        self, display_name: Optional[str] = None
    ) -> models.ChartInfoClass:
        """Ensure chart properties exist, using a safer approach."""
        return self._setdefault_aspect(
            models.ChartInfoClass(
                title=display_name or self.urn.chart_id,
                description="",
                lastModified=models.ChangeAuditStampsClass(),
            )
        )

    def _get_audit_stamps(self) -> models.ChangeAuditStampsClass:
        """Get the audit stamps from the chart properties."""
        return self._ensure_chart_props().lastModified

    def _set_audit_stamps(self, audit_stamps: models.ChangeAuditStampsClass) -> None:
        """Set the audit stamps on the chart properties."""
        self._ensure_chart_props().lastModified = audit_stamps

    @property
    def name(self) -> str:
        """Get the name of the chart."""
        return self.urn.chart_id

    @property
    @deprecated("Use display_name instead", version="1.2.0.7")
    def title(self) -> str:
        """Get the display name of the chart."""
        return self.display_name

    @deprecated("Use set_display_name instead", version="1.2.0.7")
    def set_title(self, title: str) -> None:
        """Set the display name of the chart."""
        self.set_display_name(title)

    @property
    def description(self) -> Optional[str]:
        """Get the description of the chart."""
        return self._ensure_chart_props().description

    def set_description(self, description: str) -> None:
        """Set the description of the chart."""
        self._ensure_chart_props().description = description

    @property
    def display_name(self) -> str:
        """Get the display name of the chart."""
        return self._ensure_chart_props().title

    def set_display_name(self, display_name: str) -> None:
        """Set the display name of the chart."""
        self._ensure_chart_props().title = display_name

    @property
    def external_url(self) -> Optional[str]:
        """Get the external URL of the chart."""
        return self._ensure_chart_props().externalUrl

    def set_external_url(self, external_url: str) -> None:
        """Set the external URL of the chart."""
        self._ensure_chart_props().externalUrl = external_url

    @property
    def chart_url(self) -> Optional[str]:
        """Get the chart URL."""
        return self._ensure_chart_props().chartUrl

    def set_chart_url(self, chart_url: str) -> None:
        """Set the chart URL."""
        self._ensure_chart_props().chartUrl = chart_url

    @property
    def custom_properties(self) -> Dict[str, str]:
        """Get the custom properties of the chart."""
        return self._ensure_chart_props().customProperties

    def set_custom_properties(self, custom_properties: Dict[str, str]) -> None:
        """Set the custom properties of the chart."""
        self._ensure_chart_props().customProperties = custom_properties

    @property
    def last_refreshed(self) -> Optional[datetime]:
        """Get the last refresh timestamp of the chart."""
        last_refreshed_time = self._ensure_chart_props().lastRefreshed
        return (
            datetime.fromtimestamp(last_refreshed_time)
            if last_refreshed_time is not None
            else None
        )

    def set_last_refreshed(self, last_refreshed: datetime) -> None:
        """Set the last refresh timestamp of the chart."""
        chart_props = self._ensure_chart_props()
        chart_props.lastRefreshed = int(last_refreshed.timestamp())

    @property
    def chart_type(self) -> Optional[str]:
        """Get the type of the chart as a string."""
        chart_type = self._ensure_chart_props().type
        return str(chart_type) if chart_type is not None else None

    def set_chart_type(self, chart_type: Union[str, models.ChartTypeClass]) -> None:
        """Set the type of the chart."""
        if isinstance(chart_type, str):
            chart_type_options = get_enum_options(models.ChartTypeClass)
            if chart_type not in chart_type_options:
                raise ValueError(
                    f"Invalid chart type: {chart_type}; valid types are {chart_type_options}"
                )
        self._ensure_chart_props().type = chart_type

    @property
    def access(self) -> Optional[str]:
        """Get the access level of the chart as a string."""
        access = self._ensure_chart_props().access
        return str(access) if access is not None else None

    def set_access(self, access: Union[str, models.AccessLevelClass]) -> None:
        """Set the access level of the chart."""
        if isinstance(access, str):
            assert access in get_enum_options(models.AccessLevelClass), (
                f"Invalid access level: {access}"
            )
        self._ensure_chart_props().access = access

    @property
    def input_datasets(self) -> List[DatasetUrn]:
        """Get the input datasets of the chart."""
        props = self._ensure_chart_props()
        # Convert all inputs to DatasetUrn
        return [DatasetUrn.from_string(input_urn) for input_urn in (props.inputs or [])]

    def set_input_datasets(
        self, input_datasets: Sequence[Union[DatasetUrnOrStr, Dataset]]
    ) -> None:
        """Set the input datasets of the chart."""
        # Convert all inputs to strings
        inputs = []
        for input_dataset in input_datasets:
            if isinstance(input_dataset, Dataset):
                inputs.append(str(input_dataset.urn))
            else:
                inputs.append(str(input_dataset))
        self._ensure_chart_props().inputs = inputs

    def add_input_dataset(self, input_dataset: Union[DatasetUrnOrStr, Dataset]) -> None:
        """Add an input to the chart."""
        if isinstance(input_dataset, Dataset):
            input_dataset_urn = input_dataset.urn
        elif isinstance(input_dataset, str):
            input_dataset_urn = DatasetUrn.from_string(input_dataset)
        else:  # isinstance(input_dataset, DatasetUrn)
            input_dataset_urn = input_dataset

        chart_props = self._ensure_chart_props()
        inputs = chart_props.inputs or []
        if str(input_dataset_urn) not in inputs:
            inputs.append(str(input_dataset_urn))
        chart_props.inputs = inputs

    def remove_input_dataset(
        self, input_dataset: Union[DatasetUrnOrStr, Dataset]
    ) -> None:
        """Remove an input from the chart."""
        chart_props = self._ensure_chart_props()
        inputs = chart_props.inputs or []
        if input_dataset in inputs:
            inputs.remove(str(input_dataset))
            chart_props.inputs = inputs



import re
from abc import ABC, abstractmethod

# Add imports for source customization
from typing import Any, Callable, Dict, Optional, Type, TypeVar
from urllib.parse import unquote

# Don't import TableData at the module level to avoid circular imports
# from datahub.ingestion.source.s3.source import TableData

T = TypeVar("T")


class ObjectStoreInterface(ABC):
    """
    Abstract interface for object store operations.

    This interface defines the operations that any object store connector
    (S3, GCS, ABS, etc.) should implement to provide a consistent API.
    """

    @classmethod
    @abstractmethod
    def is_uri(cls, uri: str) -> bool:
        """
        Check if the given URI is for this object store.

        Args:
            uri: The URI to check

        Returns:
            True if the URI is for this object store, False otherwise
        """
        pass

    @classmethod
    @abstractmethod
    def get_prefix(cls, uri: str) -> Optional[str]:
        """
        Get the prefix for this object store URI (e.g., 's3://', 'gs://').

        Args:
            uri: The URI to get the prefix from

        Returns:
            The prefix if the URI starts with it, None otherwise
        """
        pass

    @classmethod
    @abstractmethod
    def strip_prefix(cls, uri: str) -> str:
        """
        Remove the object store prefix from the URI.

        Args:
            uri: The URI to strip the prefix from

        Returns:
            The URI without the prefix

        Raises:
            ValueError: If the URI does not start with the expected prefix
        """
        pass

    @classmethod
    @abstractmethod
    def get_bucket_name(cls, uri: str) -> str:
        """
        Get the bucket name from the URI.

        Args:
            uri: The URI to get the bucket name from

        Returns:
            The bucket name

        Raises:
            ValueError: If the URI is not valid for this object store
        """
        pass

    @classmethod
    @abstractmethod
    def get_object_key(cls, uri: str) -> str:
        """
        Get the object key/path (excluding the bucket) from the URI.

        Args:
            uri: The URI to get the object key from

        Returns:
            The object key

        Raises:
            ValueError: If the URI is not valid for this object store
        """
        pass

    @classmethod
    def get_object_store_bucket_name(cls, uri: str) -> str:
        """
        Get the bucket name from the URI, handling foreign URIs if supported.

        The default implementation just calls get_bucket_name, but subclasses
        can override this to handle URIs from other object stores.

        Args:
            uri: The URI to get the bucket name from

        Returns:
            The bucket name

        Raises:
            ValueError: If the URI is not supported
        """
        return cls.get_bucket_name(uri)


class S3ObjectStore(ObjectStoreInterface):
    """Implementation of ObjectStoreInterface for Amazon S3."""

    PREFIXES = ["s3://", "s3n://", "s3a://"]

    @classmethod
    def is_uri(cls, uri: str) -> bool:
        return any(uri.startswith(prefix) for prefix in cls.PREFIXES)

    @classmethod
    def get_prefix(cls, uri: str) -> Optional[str]:
        for prefix in cls.PREFIXES:
            if uri.startswith(prefix):
                return prefix
        return None

    @classmethod
    def strip_prefix(cls, uri: str) -> str:
        prefix = cls.get_prefix(uri)
        if not prefix:
            raise ValueError(
                f"Not an S3 URI. Must start with one of the following prefixes: {str(cls.PREFIXES)}"
            )
        return uri[len(prefix) :]

    @classmethod
    def get_bucket_name(cls, uri: str) -> str:
        if not cls.is_uri(uri):
            raise ValueError(
                f"Not an S3 URI. Must start with one of the following prefixes: {str(cls.PREFIXES)}"
            )
        return cls.strip_prefix(uri).split("/")[0]

    @classmethod
    def get_object_key(cls, uri: str) -> str:
        if not cls.is_uri(uri):
            raise ValueError(
                f"Not an S3 URI. Must start with one of the following prefixes: {str(cls.PREFIXES)}"
            )
        parts = cls.strip_prefix(uri).split("/", 1)
        if len(parts) < 2:
            return ""
        return parts[1]

    @classmethod
    def get_object_store_bucket_name(cls, uri: str) -> str:
        """
        Get the bucket name from an S3 URI.

        Args:
            uri: The URI to get the bucket name from

        Returns:
            The bucket name

        Raises:
            ValueError: If the URI is not an S3 URI
        """
        return cls.get_bucket_name(uri)


class GCSObjectStore(ObjectStoreInterface):
    """Implementation of ObjectStoreInterface for Google Cloud Storage."""

    PREFIX = "gs://"

    @classmethod
    def is_uri(cls, uri: str) -> bool:
        return uri.startswith(cls.PREFIX)

    @classmethod
    def get_prefix(cls, uri: str) -> Optional[str]:
        if uri.startswith(cls.PREFIX):
            return cls.PREFIX
        return None

    @classmethod
    def strip_prefix(cls, uri: str) -> str:
        prefix = cls.get_prefix(uri)
        if not prefix:
            raise ValueError(f"Not a GCS URI. Must start with prefix: {cls.PREFIX}")
        return uri[len(prefix) :]

    @classmethod
    def get_bucket_name(cls, uri: str) -> str:
        if not cls.is_uri(uri):
            raise ValueError(f"Not a GCS URI. Must start with prefix: {cls.PREFIX}")
        return cls.strip_prefix(uri).split("/")[0]

    @classmethod
    def get_object_key(cls, uri: str) -> str:
        if not cls.is_uri(uri):
            raise ValueError(f"Not a GCS URI. Must start with prefix: {cls.PREFIX}")
        parts = cls.strip_prefix(uri).split("/", 1)
        if len(parts) < 2:
            return ""
        return parts[1]

    @classmethod
    def get_object_store_bucket_name(cls, uri: str) -> str:
        """
        Get the bucket name from a GCS URI.

        Args:
            uri: The URI to get the bucket name from

        Returns:
            The bucket name

        Raises:
            ValueError: If the URI is not a GCS URI
        """
        return cls.get_bucket_name(uri)


class ABSObjectStore(ObjectStoreInterface):
    """Implementation of ObjectStoreInterface for Azure Blob Storage."""

    PREFIX = "abfss://"
    HTTPS_REGEX = re.compile(r"(https?://[a-z0-9]{3,24}\.blob\.core\.windows\.net/)")

    @classmethod
    def is_uri(cls, uri: str) -> bool:
        return uri.startswith(cls.PREFIX) or bool(cls.HTTPS_REGEX.match(uri))

    @classmethod
    def get_prefix(cls, uri: str) -> Optional[str]:
        if uri.startswith(cls.PREFIX):
            return cls.PREFIX

        # Check for HTTPS format
        match = cls.HTTPS_REGEX.match(uri)
        if match:
            return match.group(1)

        return None

    @classmethod
    def strip_prefix(cls, uri: str) -> str:
        if uri.startswith(cls.PREFIX):
            return uri[len(cls.PREFIX) :]

        # Handle HTTPS format
        match = cls.HTTPS_REGEX.match(uri)
        if match:
            return uri[len(match.group(1)) :]

        raise ValueError(
            f"Not an ABS URI. Must start with prefix: {cls.PREFIX} or match Azure Blob Storage HTTPS pattern"
        )

    @classmethod
    def get_bucket_name(cls, uri: str) -> str:
        if not cls.is_uri(uri):
            raise ValueError(
                f"Not an ABS URI. Must start with prefix: {cls.PREFIX} or match Azure Blob Storage HTTPS pattern"
            )

        if uri.startswith(cls.PREFIX):
            # abfss://container@account.dfs.core.windows.net/path
            return cls.strip_prefix(uri).split("@")[0]
        else:
            # https://account.blob.core.windows.net/container/path
            return cls.strip_prefix(uri).split("/")[0]

    @classmethod
    def get_object_key(cls, uri: str) -> str:
        if not cls.is_uri(uri):
            raise ValueError(
                f"Not an ABS URI. Must start with prefix: {cls.PREFIX} or match Azure Blob Storage HTTPS pattern"
            )

        if uri.startswith(cls.PREFIX):
            # abfss://container@account.dfs.core.windows.net/path
            parts = cls.strip_prefix(uri).split("@", 1)
            if len(parts) < 2:
                return ""
            account_path = parts[1]
            path_parts = account_path.split("/", 1)
            if len(path_parts) < 2:
                return ""
            return path_parts[1]
        else:
            # https://account.blob.core.windows.net/container/path
            stripped = cls.strip_prefix(uri)
            parts = stripped.split("/", 1)
            if len(parts) < 2:
                return ""
            return parts[1]


# Registry of all object store implementations
OBJECT_STORE_REGISTRY: Dict[str, Type[ObjectStoreInterface]] = {
    "s3": S3ObjectStore,
    "gcs": GCSObjectStore,
    "abs": ABSObjectStore,
}


def get_object_store_for_uri(uri: str) -> Optional[Type[ObjectStoreInterface]]:
    """
    Get the appropriate object store implementation for the given URI.

    Args:
        uri: The URI to get the object store for

    Returns:
        The object store implementation, or None if no matching implementation is found
    """
    for object_store in OBJECT_STORE_REGISTRY.values():
        if object_store.is_uri(uri):
            return object_store
    return None


def get_object_store_bucket_name(uri: str) -> str:
    """
    Get the bucket name from any supported object store URI.

    This function acts as a central dispatcher that:
    1. Identifies the appropriate object store implementation for the URI
    2. Uses that implementation to extract the bucket name
    3. Falls back to specific URI format parsing if needed

    Args:
        uri: The URI to get the bucket name from

    Returns:
        The bucket name

    Raises:
        ValueError: If the URI is not supported by any registered object store
    """
    # First try to find the native implementation for this URI
    object_store = get_object_store_for_uri(uri)
    if object_store:
        return object_store.get_bucket_name(uri)

    # If no native implementation, handle specific URI formats directly
    if uri.startswith("gs://"):
        return uri[5:].split("/")[0]
    elif any(uri.startswith(prefix) for prefix in S3ObjectStore.PREFIXES):
        prefix_length = next(
            len(prefix) for prefix in S3ObjectStore.PREFIXES if uri.startswith(prefix)
        )
        return uri[prefix_length:].split("/")[0]
    elif uri.startswith(ABSObjectStore.PREFIX):
        return uri[len(ABSObjectStore.PREFIX) :].split("@")[0]
    elif ABSObjectStore.HTTPS_REGEX.match(uri):
        # Handle HTTPS Azure Blob Storage URLs
        match = ABSObjectStore.HTTPS_REGEX.match(uri)
        if match:
            stripped = uri[len(match.group(1)) :]
            return stripped.split("/")[0]

    raise ValueError(f"Unsupported URI format: {uri}")


def get_object_key(uri: str) -> str:
    """
    Get the object key from any supported object store URI.

    Args:
        uri: The URI to get the object key from

    Returns:
        The object key

    Raises:
        ValueError: If the URI is not supported by any registered object store
    """
    object_store = get_object_store_for_uri(uri)
    if object_store:
        return object_store.get_object_key(uri)

    raise ValueError(f"Unsupported URI format: {uri}")


class ObjectStoreSourceAdapter:
    """
    Adapter for customizing object store source implementations.

    This class provides a way to customize source implementations for different
    object stores (S3, GCS, etc.) without having to directly modify those classes.
    Instead, adapters register customizations that are applied to the source instance.
    """

    @staticmethod
    def create_s3_path(bucket_name: str, key: str) -> str:
        """
        Create a default S3 path.

        Args:
            bucket_name: The bucket name
            key: The object key

        Returns:
            A properly formatted S3 URI
        """
        return unquote(f"s3://{bucket_name}/{key}")

    @staticmethod
    def create_gcs_path(bucket_name: str, key: str) -> str:
        """
        Create a default GCS path.

        Args:
            bucket_name: The bucket name
            key: The object key

        Returns:
            A properly formatted GCS URI
        """
        return unquote(f"gs://{bucket_name}/{key}")

    @staticmethod
    def create_abs_path(container_name: str, key: str, account_name: str) -> str:
        """
        Create a default Azure Blob Storage path.

        Args:
            container_name: The container name
            key: The object key
            account_name: The storage account name

        Returns:
            A properly formatted ABS URI
        """
        return unquote(
            f"abfss://{container_name}@{account_name}.dfs.core.windows.net/{key}"
        )

    @staticmethod
    def get_s3_external_url(
        table_data: Any, region: Optional[str] = None
    ) -> Optional[str]:
        """
        Get the AWS S3 console URL for the given table.

        Args:
            table_data: Table data containing path information
            region: AWS region for the S3 console URL, defaults to us-east-1 if not specified

        Returns:
            The AWS console URL, or None if not applicable
        """
        if not S3ObjectStore.is_uri(table_data.table_path):
            return None

        # Get the bucket name and key from the S3 URI
        bucket_name = get_object_store_bucket_name(table_data.table_path)
        key = get_object_key(table_data.table_path)

        # Use the provided region or default to us-east-1
        aws_region = region or "us-east-1"

        return f"https://{aws_region}.console.aws.amazon.com/s3/buckets/{bucket_name}?prefix={key}"

    @staticmethod
    def get_gcs_external_url(table_data: Any) -> Optional[str]:
        """
        Get the GCS console URL for the given table.

        Args:
            table_data: Table data containing path information

        Returns:
            The GCS console URL, or None if not applicable
        """
        if not GCSObjectStore.is_uri(table_data.table_path):
            return None

        # Get the bucket name and key from the GCS URI
        bucket_name = get_object_store_bucket_name(table_data.table_path)
        key = get_object_key(table_data.table_path)

        # Return the basic GCS console URL
        return f"https://console.cloud.google.com/storage/browser/{bucket_name}/{key}"

    @staticmethod
    def get_abs_external_url(table_data: Any) -> Optional[str]:
        """
        Get the Azure Storage browser URL for the given table.

        Args:
            table_data: Table data containing path information

        Returns:
            The Azure Storage URL, or None if not applicable
        """
        if not ABSObjectStore.is_uri(table_data.table_path):
            return None

        try:
            if table_data.table_path.startswith("abfss://"):
                # URI format: abfss://container@account.dfs.core.windows.net/path
                path_without_prefix = ABSObjectStore.strip_prefix(table_data.table_path)
                parts = path_without_prefix.split("@", 1)
                if len(parts) < 2:
                    return None

                container_name = parts[0]
                account_parts = parts[1].split("/", 1)
                account_domain = account_parts[0]
                account_name = account_domain.split(".")[0]
            else:
                # Handle HTTPS format: https://account.blob.core.windows.net/container/path
                container_name = ABSObjectStore.get_bucket_name(table_data.table_path)
                if "blob.core.windows.net" in table_data.table_path:
                    account_name = table_data.table_path.split("//")[1].split(".")[0]
                else:
                    return None

            # Construct Azure portal URL
            return f"https://portal.azure.com/#blade/Microsoft_Azure_Storage/ContainerMenuBlade/overview/storageAccountId/{account_name}/containerName/{container_name}"
        except Exception:
            # If any parsing error occurs, return None
            return None

    def __init__(
        self,
        platform: str,
        platform_name: str,
        aws_region: Optional[str] = None,
        azure_storage_account: Optional[str] = None,
    ):
        """
        Initialize the adapter with platform-specific configurations.

        Args:
            platform: The platform identifier (e.g., "s3", "gcs", "abs")
            platform_name: The human-readable platform name
            aws_region: AWS region for S3 URLs, defaults to us-east-1 if not specified
            azure_storage_account: Azure Storage account name
        """
        self.platform = platform
        self.platform_name = platform_name
        self.aws_region = aws_region
        self.azure_storage_account = azure_storage_account
        self.customizations: Dict[str, Callable[..., Any]] = {}

        # Register default customizations based on platform
        if platform == "gcs":
            self.register_customization("is_s3_platform", lambda: True)
            self.register_customization("create_s3_path", self.create_gcs_path)
            self.register_customization(
                "get_external_url",
                lambda table_data: self.get_gcs_external_url(table_data),
            )
            # Fix URI mismatch issue in pattern matching
            self.register_customization(
                "_normalize_uri_for_pattern_matching",
                self._normalize_gcs_uri_for_pattern_matching,
            )
            # Fix URI handling in schema extraction - override strip_s3_prefix for GCS
            self.register_customization("strip_s3_prefix", self._strip_gcs_prefix)
        elif platform == "s3":
            self.register_customization("is_s3_platform", lambda: True)
            self.register_customization("create_s3_path", self.create_s3_path)
            self.register_customization(
                "get_external_url",
                lambda table_data: self.get_s3_external_url(
                    table_data, self.aws_region
                ),
            )
        elif platform == "abs":
            self.register_customization("is_s3_platform", lambda: True)
            # If we have an Azure storage account, create a specialized path creation function
            if self.azure_storage_account:
                storage_account = (
                    self.azure_storage_account
                )  # Create a local non-optional variable
                self.register_customization(
                    "create_s3_path",
                    lambda bucket, key: self.create_abs_path(
                        bucket, key, storage_account
                    ),
                )
            else:
                # Fall back to a simpler implementation if no account provided
                self.register_customization(
                    "create_s3_path", lambda bucket, key: f"abfss://{bucket}@{key}"
                )
            self.register_customization("get_external_url", self.get_abs_external_url)

    def register_customization(
        self, method_name: str, implementation: Callable[..., Any]
    ) -> None:
        """
        Register a customization for a specific method.

        Args:
            method_name: The name of the method to customize
            implementation: The implementation to use
        """
        self.customizations[method_name] = implementation

    def apply_customizations(self, source: Any) -> Any:
        """
        Apply all registered customizations to the source instance.

        Args:
            source: The source instance to customize

        Returns:
            The customized source instance
        """
        # Set the platform
        if hasattr(source, "source_config") and hasattr(
            source.source_config, "platform"
        ):
            source.source_config.platform = self.platform

        # Apply method customizations
        for method_name, implementation in self.customizations.items():
            # For instance methods that use self, we need to bind them to the source
            if (
                hasattr(implementation, "__self__")
                and implementation.__self__ is not None
            ):
                # This is already a bound method, use __get__ to rebind it to the source
                setattr(source, method_name, implementation.__get__(source))
            else:
                # This is a regular function or static/class method
                setattr(source, method_name, implementation)

        return source

    # Add a direct method for tests that may call this directly
    def get_external_url(self, table_data: Any) -> Optional[str]:
        """
        Get the external URL for a table based on the platform type.

        This method routes to the appropriate implementation based on the platform.

        Args:
            table_data: Table data containing path information

        Returns:
            An external URL or None if not applicable
        """
        if self.platform == "s3":
            return self.get_s3_external_url(table_data, self.aws_region)
        elif self.platform == "gcs":
            return self.get_gcs_external_url(table_data)
        elif self.platform == "abs":
            return self.get_abs_external_url(table_data)
        return None

    def _normalize_gcs_uri_for_pattern_matching(self, uri: str) -> str:
        """
        Normalize GCS URI for pattern matching.

        This method converts gs:// URIs to s3:// URIs for pattern matching purposes,
        fixing the URI mismatch issue in GCS ingestion.

        Args:
            uri: The URI to normalize

        Returns:
            The normalized URI for pattern matching
        """
        if uri.startswith("gs://"):
            return uri.replace("gs://", "s3://", 1)
        return uri

    def _strip_gcs_prefix(self, uri: str) -> str:
        """
        Strip GCS prefix from URI.

        This method removes the gs:// prefix from GCS URIs for path processing.

        Args:
            uri: The URI to strip the prefix from

        Returns:
            The URI without the gs:// prefix
        """
        if uri.startswith("gs://"):
            return uri[5:]  # Remove "gs://" prefix
        return uri


# Factory function to create an adapter for a specific platform
def create_object_store_adapter(
    platform: str,
    aws_region: Optional[str] = None,
    azure_storage_account: Optional[str] = None,
) -> ObjectStoreSourceAdapter:
    """
    Create an adapter for a specific object store platform.

    Args:
        platform: The platform identifier (e.g., "s3", "gcs", "abs")
        aws_region: AWS region for S3 URLs, defaults to us-east-1 if not specified
        azure_storage_account: Azure Storage account name

    Returns:
        An adapter configured for the specified platform
    """
    platform_names = {
        "s3": "Amazon S3",
        "gcs": "Google Cloud Storage",
        "abs": "Azure Blob Storage",
    }

    return ObjectStoreSourceAdapter(
        platform=platform,
        platform_name=platform_names.get(platform, f"Unknown ({platform})"),
        aws_region=aws_region,
        azure_storage_account=azure_storage_account,
    )



"""Databricks-specific profiling adapter."""

import logging
from typing import Any, List, Optional

import sqlalchemy as sa
from sqlalchemy.engine import Connection
from sqlalchemy.sql.elements import ColumnElement

from datahub.ingestion.source.sqlalchemy_profiler.base_adapter import (
    DEFAULT_QUANTILES,
    PlatformAdapter,
)

logger = logging.getLogger(__name__)


class DatabricksAdapter(PlatformAdapter):
    """
    Databricks-specific profiling adapter.

    Databricks optimizations:
    1. approx_count_distinct for fast unique counts
    2. approx_percentile for median calculation

    Note: Databricks uses lowercase function names (approx_count_distinct, approx_percentile)
    unlike some other platforms.

    Uses default setup_profiling and cleanup from PlatformAdapter.
    """

    # =========================================================================
    # SQL Expression Builders
    # =========================================================================

    def get_approx_unique_count_expr(self, column: str) -> ColumnElement[Any]:
        """
        Databricks uses approx_count_distinct for fast unique counts.

        This matches GE profiler behavior (ge_data_profiler.py:233-239).
        Note: Databricks uses lowercase function name.

        Args:
            column: Column name

        Returns:
            SQLAlchemy expression for approx_count_distinct
        """
        return sa.func.approx_count_distinct(sa.column(column))

    def get_median_expr(self, column: str) -> Optional[ColumnElement[Any]]:
        """
        Databricks uses approx_percentile for median.

        This matches GE profiler behavior (ge_data_profiler.py:684-693).
        approx_percentile(column, 0.5) computes the approximate median.

        Args:
            column: Column name

        Returns:
            SQLAlchemy expression for approx_percentile(column, 0.5)
        """
        return sa.func.approx_percentile(sa.column(column), 0.5)

    def get_column_quantiles(
        self,
        table: sa.Table,
        column: str,
        conn: Connection,
        quantiles: Optional[List[float]] = None,
    ) -> List[Optional[float]]:
        """
        Get quantile values for a column using Databricks' approx_percentile.

        Databricks: approx_percentile(col, array(0.05, 0.25, ...)) returns an array.

        Args:
            table: SQLAlchemy table object
            column: Column name
            conn: Active database connection
            quantiles: List of quantile values (default: DEFAULT_QUANTILES)

        Returns:
            List of quantile values (None for unavailable quantiles)
        """
        if quantiles is None:
            quantiles = DEFAULT_QUANTILES

        quoted_column = self.quote_identifier(column)
        # Databricks: Similar to Athena/Trino but uses array() syntax
        array_str = f"array({', '.join(str(q) for q in quantiles)})"
        databricks_expr = sa.literal_column(
            f"approx_percentile({quoted_column}, {array_str})"
        ).label("quantiles")
        query = sa.select([databricks_expr]).select_from(table)
        result = conn.execute(query).scalar()
        logger.debug(
            f"Databricks quantiles for {column}: result type={type(result)}, "
            f"value={result}, expected_length={len(quantiles)}"
        )
        # Result is an array, convert to list
        if isinstance(result, list):
            if len(result) != len(quantiles):
                logger.warning(
                    f"Quantile result length mismatch: got {len(result)}, expected {len(quantiles)}"
                )
            return [float(v) if v is not None else None for v in result]
        logger.warning(
            f"Quantile result is not a list: type={type(result)}, value={result}"
        )
        return [None] * len(quantiles)



from typing import Any, TextIO


class TeeIO:
    """
    A file-like object that writes to multiple streams, similar to `tee`.
    It mirrors the attributes of the first stream for encoding/tty/etc.
    """

    def __init__(self, *args: TextIO) -> None:
        assert args
        self._streams = args

    def write(self, line: str) -> None:
        for stream in self._streams:
            stream.write(line)

    def flush(self) -> None:
        for stream in self._streams:
            stream.flush()

    def __getattr__(self, attr: str) -> Any:
        return getattr(self._streams[0], attr)



import json
import logging
import os.path
import re
from typing import Dict, Optional

import click
import packaging
import requests
import yaml
from packaging.version import parse
from pydantic import BaseModel

from datahub._version import nice_version_name
from datahub.configuration.env_vars import get_force_local_quickstart_mapping

logger = logging.getLogger(__name__)

LOCAL_QUICKSTART_MAPPING_FILE = get_force_local_quickstart_mapping()
DEFAULT_LOCAL_CONFIG_PATH = "~/.datahub/quickstart/quickstart_version_mapping.yaml"
DEFAULT_REMOTE_CONFIG_PATH = "https://raw.githubusercontent.com/datahub-project/datahub/master/docker/quickstart/quickstart_version_mapping.yaml"

MINIMUM_SUPPORTED_VERSION = "v1.1.0"


def get_minimum_supported_version_message(version: str) -> str:
    MINIMUM_SUPPORTED_VERSION_MESSAGE = f"""
    DataHub CLI Version Compatibility Issue

    You're trying to install DataHub server version {version} which is not supported by this CLI version.

    This CLI (version {nice_version_name()}) only supports installing DataHub server versions {MINIMUM_SUPPORTED_VERSION} and above.

    To install older server versions:
    1. Uninstall current CLI: pip uninstall acryl-datahub
    2. Install older CLI: pip install acryl-datahub==1.1
    3. Run quickstart with your desired version: datahub docker quickstart --version <version>

    For more information: https://docs.datahub.com/docs/quickstart#install-datahub-server
    """
    return MINIMUM_SUPPORTED_VERSION_MESSAGE


class QuickstartExecutionPlan(BaseModel):
    composefile_git_ref: str
    docker_tag: str
    mysql_tag: Optional[str] = None


def _is_it_a_version(version: str) -> bool:
    """
    Checks if a string is a valid version.
    :param version: The string to check.
    :return: True if the string is a valid version, False otherwise.
    """
    return re.match(r"^v?\d+\.\d+(\.\d+)?$", version) is not None


class QuickstartVersionMappingConfig(BaseModel):
    quickstart_version_map: Dict[str, QuickstartExecutionPlan]

    @classmethod
    def _fetch_latest_version(cls) -> str:
        """
        Fetches the latest version from github.
        :return: The latest version.
        """
        response = requests.get(
            "https://api.github.com/repos/datahub-project/datahub/releases/latest"
        )
        response.raise_for_status()
        return json.loads(response.text)["tag_name"]

    @classmethod
    def fetch_quickstart_config(cls) -> "QuickstartVersionMappingConfig":
        if LOCAL_QUICKSTART_MAPPING_FILE:
            logger.info(
                "LOCAL_QUICKSTART_MAPPING_FILE is set, will try to read from local file."
            )
            path = os.path.expanduser(LOCAL_QUICKSTART_MAPPING_FILE)
            with open(path) as f:
                config_raw = yaml.safe_load(f)
            return cls.model_validate(config_raw)

        config_raw = None
        try:
            response = requests.get(DEFAULT_REMOTE_CONFIG_PATH, timeout=5)
            response.raise_for_status()
            config_raw = yaml.safe_load(response.text)
        except Exception as e:
            logger.debug(
                f"Couldn't connect to github: {e}, will try to read from local file."
            )
            try:
                path = os.path.expanduser(DEFAULT_LOCAL_CONFIG_PATH)
                with open(path) as f:
                    config_raw = yaml.safe_load(f)
            except Exception:
                logger.debug("Couldn't read from local file either.")

        if config_raw is None:
            logger.info(
                "Unable to connect to GitHub, using default quickstart version mapping config."
            )
            return QuickstartVersionMappingConfig(
                quickstart_version_map={
                    "default": QuickstartExecutionPlan(
                        composefile_git_ref="master", docker_tag="head", mysql_tag="8.2"
                    ),
                }
            )

        config = cls.model_validate(config_raw)

        # If stable is not defined in the config, we need to fetch the latest version from github.
        if config.quickstart_version_map.get("stable") is None:
            try:
                release = cls._fetch_latest_version()
                config.quickstart_version_map["stable"] = QuickstartExecutionPlan(
                    composefile_git_ref=release, docker_tag=release, mysql_tag="8.2"
                )
            except Exception:
                click.echo(
                    "Couldn't connect to github. --version stable will not work."
                )
        save_quickstart_config(config)
        return config

    def get_quickstart_execution_plan(
        self,
        requested_version: Optional[str],
    ) -> QuickstartExecutionPlan:
        """
        From the requested version and stable flag, returns the execution plan for the quickstart.
        Including the docker tag, composefile git ref, required containers, and checks to run.
        :return: The execution plan for the quickstart.
        """
        if requested_version is None:
            requested_version = "default"
        composefile_git_ref = requested_version
        docker_tag = requested_version
        # Default to 8.2 if not specified in version map
        mysql_tag = "8.2"
        result = self.quickstart_version_map.get(
            requested_version,
            QuickstartExecutionPlan(
                composefile_git_ref=composefile_git_ref,
                docker_tag=docker_tag,
                mysql_tag=str(mysql_tag),
            ),
        )

        if not is_minimum_supported_version(requested_version):
            click.secho(
                get_minimum_supported_version_message(version=requested_version),
                fg="red",
            )
            raise click.ClickException("Minimum supported version not met")

        # new CLI version is downloading the composefile corresponding to the requested version
        # if the version is older than <MINIMUM_SUPPORTED_VERSION>, it doesn't contain the
        # docker compose based resolved compose file. In those cases, we pick up the composefile from
        # MINIMUM_SUPPORTED_VERSION which contains the compose file.
        if _is_it_a_version(result.composefile_git_ref):
            if parse("v1.2.0") > parse(result.composefile_git_ref):
                # The merge commit where profiles based resolved compose file was added.
                # https://github.com/datahub-project/datahub/pull/13566
                result.composefile_git_ref = "21726bc3341490f4182b904626c793091ac95edd"

        return result


def save_quickstart_config(
    config: QuickstartVersionMappingConfig, path: str = DEFAULT_LOCAL_CONFIG_PATH
) -> None:
    # create directory if it doesn't exist
    path = os.path.expanduser(path)
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "w") as f:
        yaml.dump(config.model_dump(), f)
    logger.info(f"Saved quickstart config to {path}.")


def is_minimum_supported_version(version: str) -> bool:
    if not _is_it_a_version(version):
        return True

    requested_version = packaging.version.parse(version)
    minimum_supported_version = packaging.version.parse(MINIMUM_SUPPORTED_VERSION)
    if requested_version < minimum_supported_version:
        return False

    return True



"""
Source that extracts metadata from Snowplow BDP Console API or Iglu Registry.

Implementation notes:
- Uses dedicated processor classes for each entity type (schemas, event specs, tracking plans)
- Supports both BDP API and open-source Iglu registry via mode selection
- Optional warehouse lineage extraction via SQL query parsing against atomic events table
- Implements caching for schema resolution and API responses
"""

import logging
from typing import Any, Dict, Iterable, Optional

from datahub.emitter.mcp_builder import gen_containers
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import (
    CapabilityReport,
    SourceCapability,
    TestableSource,
    TestConnectionReport,
)
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.common.subtypes import DatasetContainerSubTypes
from datahub.ingestion.source.snowplow.builders.container_keys import (
    SnowplowOrganizationKey,
)
from datahub.ingestion.source.snowplow.builders.lineage_builder import LineageBuilder
from datahub.ingestion.source.snowplow.builders.ownership_builder import (
    OwnershipBuilder,
)
from datahub.ingestion.source.snowplow.builders.urn_factory import SnowplowURNFactory
from datahub.ingestion.source.snowplow.constants import (
    PLATFORM_NAME,
    ConnectionMode,
)
from datahub.ingestion.source.snowplow.dependencies import (
    IngestionState,
    ProcessorDependencies,
)
from datahub.ingestion.source.snowplow.models.snowplow_models import (
    Enrichment,
)
from datahub.ingestion.source.snowplow.processors.event_spec_processor import (
    EventSpecProcessor,
)
from datahub.ingestion.source.snowplow.processors.pipeline_processor import (
    PipelineProcessor,
)
from datahub.ingestion.source.snowplow.processors.schema_processor import (
    SchemaProcessor,
)
from datahub.ingestion.source.snowplow.processors.standard_schema_processor import (
    StandardSchemaProcessor,
)
from datahub.ingestion.source.snowplow.processors.tracking_plan_processor import (
    TrackingPlanProcessor,
)
from datahub.ingestion.source.snowplow.processors.warehouse_lineage_processor import (
    WarehouseLineageProcessor,
)
from datahub.ingestion.source.snowplow.services.atomic_event_builder import (
    AtomicEventBuilder,
)
from datahub.ingestion.source.snowplow.services.column_lineage_builder import (
    ColumnLineageBuilder,
)
from datahub.ingestion.source.snowplow.services.data_structure_builder import (
    DataStructureBuilder,
)
from datahub.ingestion.source.snowplow.services.deployment_fetcher import (
    DeploymentFetcher,
)
from datahub.ingestion.source.snowplow.services.enrichment_registry_factory import (
    EnrichmentRegistryFactory,
)
from datahub.ingestion.source.snowplow.services.error_handler import ErrorHandler
from datahub.ingestion.source.snowplow.services.field_tagging import (
    FieldTagger,
)
from datahub.ingestion.source.snowplow.services.iglu_client import IgluClient
from datahub.ingestion.source.snowplow.services.property_manager import PropertyManager
from datahub.ingestion.source.snowplow.services.user_resolver import UserResolver
from datahub.ingestion.source.snowplow.snowplow_client import SnowplowBDPClient
from datahub.ingestion.source.snowplow.snowplow_config import SnowplowSourceConfig
from datahub.ingestion.source.snowplow.snowplow_report import SnowplowSourceReport
from datahub.ingestion.source.snowplow.utils.cache_manager import CacheManager
from datahub.ingestion.source.snowplow.utils.field_reference_parser import (
    FieldReferenceParser,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionSourceBase,
)
from datahub.utilities.registries.domain_registry import DomainRegistry

logger = logging.getLogger(__name__)


@platform_name("Snowplow")
@config_class(SnowplowSourceConfig)
@support_status(SupportStatus.INCUBATING)
@capability(SourceCapability.PLATFORM_INSTANCE, "Enabled by default")
@capability(SourceCapability.DOMAINS, "Supported via configuration")
@capability(
    SourceCapability.SCHEMA_METADATA,
    "Enabled by default for event and entity schemas",
)
@capability(
    SourceCapability.DESCRIPTIONS, "Enabled by default from schema descriptions"
)
@capability(
    SourceCapability.LINEAGE_COARSE,
    "Optionally enabled via warehouse_lineage.enabled configuration (requires BDP)",
)
@capability(SourceCapability.DELETION_DETECTION, "Enabled via stateful ingestion")
class SnowplowSource(StatefulIngestionSourceBase, TestableSource):
    """
    Ingests metadata from Snowplow.

    Extracts:
    - Organizations (as containers)
    - Event schemas (as datasets)
    - Entity schemas (as datasets)
    - Event specifications (as datasets) - BDP only
    - Tracking scenarios (as containers) - BDP only
    - Warehouse lineage (optional) - requires warehouse connection

    Supports:
    - Snowplow BDP (Behavioral Data Platform) deployments
    - Open-source Snowplow with Iglu registry
    """

    config: SnowplowSourceConfig
    report: SnowplowSourceReport
    platform = PLATFORM_NAME

    def __init__(self, config: SnowplowSourceConfig, ctx: PipelineContext):
        super().__init__(config, ctx)
        self.config = config
        self.report = SnowplowSourceReport()

        # Initialize API clients
        self.bdp_client: Optional[SnowplowBDPClient] = None
        self.iglu_client: Optional[IgluClient] = None

        if config.bdp_connection:
            self.bdp_client = SnowplowBDPClient(
                config.bdp_connection, report=self.report
            )
            self.report.connection_mode = ConnectionMode.BDP.value
            self.report.organization_id = config.bdp_connection.organization_id

        if config.iglu_connection:
            self.iglu_client = IgluClient(config.iglu_connection, report=self.report)
            if self.report.connection_mode == ConnectionMode.BDP.value:
                self.report.connection_mode = ConnectionMode.BOTH.value
            else:
                self.report.connection_mode = ConnectionMode.IGLU.value

        # Initialize user resolver for ownership resolution
        self.user_resolver = UserResolver(self.bdp_client, self.report)
        self.user_resolver.load_users()

        # Initialize stale entity removal handler
        self.stale_entity_removal_handler = StaleEntityRemovalHandler.create(
            self, self.config, self.ctx
        )

        # Domain registry (optional)
        self.domain_registry: Optional[DomainRegistry] = None
        if hasattr(self.config, "domain") and self.config.domain:
            self.domain_registry = DomainRegistry(
                cached_domains=[k for k in self.config.domain], graph=self.ctx.graph
            )

        # Initialize enrichment lineage registry
        self.enrichment_lineage_registry = EnrichmentRegistryFactory.create_registry()

        # Initialize field tagger
        self.field_tagger = FieldTagger(self.config.field_tagging)

        # Initialize URN factory for centralized URN construction
        self.urn_factory = SnowplowURNFactory(
            platform=self.platform,
            config=self.config,
        )

        # Initialize cache manager for performance optimization
        self.cache = CacheManager()

        # Initialize builders for metadata construction
        self.ownership_builder = OwnershipBuilder(
            config=self.config,
            user_cache=self.user_resolver.get_user_cache(),
            user_name_cache=self.user_resolver.get_user_name_cache(),
        )
        self.lineage_builder = LineageBuilder()

        # Initialize error handler for consistent error handling
        self.error_handler = ErrorHandler(report=self.report)

        # Initialize property manager for structured properties
        self.property_manager = PropertyManager(config=self.config)

        # Create mutable ingestion state (populated during extraction)
        # This is separate from deps to maintain clean separation between
        # immutable config and mutable runtime data
        self.state = IngestionState()

        # Initialize column lineage builder
        self.column_lineage_builder = ColumnLineageBuilder(
            config=self.config,
            urn_factory=self.urn_factory,
            state=self.state,
        )

        # Initialize atomic event builder
        self.atomic_event_builder = AtomicEventBuilder(
            config=self.config,
            urn_factory=self.urn_factory,
            state=self.state,
            platform=self.platform,
        )

        # Initialize deployment fetcher service
        self.deployment_fetcher = DeploymentFetcher(
            bdp_client=self.bdp_client,
            enable_parallel=self.config.performance.enable_parallel_fetching,
            max_workers=self.config.performance.max_concurrent_api_calls,
            report=self.report,
        )

        # Create processor dependencies (dependency injection)
        # Immutable dependencies only - state is managed separately
        self.deps = ProcessorDependencies(
            config=self.config,
            report=self.report,
            cache=self.cache,
            platform=self.platform,
            urn_factory=self.urn_factory,
            ownership_builder=self.ownership_builder,
            lineage_builder=self.lineage_builder,
            enrichment_lineage_registry=self.enrichment_lineage_registry,
            error_handler=self.error_handler,
            field_tagger=self.field_tagger,
            bdp_client=self.bdp_client,
            iglu_client=self.iglu_client,
            graph=self.ctx.graph if hasattr(self.ctx, "graph") else None,
            deployment_fetcher=self.deployment_fetcher,
        )

        # Initialize data structure builder (used by SchemaProcessor)
        self.data_structure_builder = DataStructureBuilder(
            deps=self.deps,
            state=self.state,
            column_lineage_builder=self.column_lineage_builder,
        )

        # Initialize processors for entity extraction
        # Processors receive both immutable deps and mutable state
        self.schema_processor = SchemaProcessor(
            deps=self.deps,
            state=self.state,
            data_structure_builder=self.data_structure_builder,
        )
        self.pipeline_processor = PipelineProcessor(deps=self.deps, state=self.state)
        self.event_spec_processor = EventSpecProcessor(deps=self.deps, state=self.state)
        self.tracking_plan_processor = TrackingPlanProcessor(
            deps=self.deps, state=self.state
        )
        self.standard_schema_processor = StandardSchemaProcessor(
            deps=self.deps, state=self.state
        )
        self.warehouse_lineage_processor = WarehouseLineageProcessor(
            deps=self.deps, state=self.state
        )

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "SnowplowSource":
        config = SnowplowSourceConfig.model_validate(config_dict)
        return cls(config, ctx)

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        """
        Main extraction logic.

        Order:
        0. Register structured property definitions (if using structured properties)
        1. Emit organization container
        2. Create atomic event dataset (synthetic schema for enrichment lineage)
        3. Extract event and entity schemas
        4. Extract event specifications (BDP only)
        5. Extract tracking scenarios (BDP only)
        6. Extract warehouse lineage (optional)
        """

        # Register structured property definitions if enabled
        # Use synchronous registration via graph client to ensure definitions are
        # committed before field property assignments are validated in the same batch
        if self.config.field_tagging.use_structured_properties:
            if self.ctx.graph:
                # Sync registration ensures definitions exist before assignments
                self.property_manager.register_structured_properties_sync(
                    self.ctx.graph
                )
            else:
                # Fall back to workunit emission (may fail with batch validation)
                logger.warning(
                    "Graph client not available for synchronous property registration. "
                    "Structured property definitions will be emitted as workunits. "
                    "If validation fails, pre-register properties with: "
                    "datahub properties upsert -f snowplow_field_structured_properties.yaml"
                )
                yield from self.property_manager.register_structured_properties()

        # Emit organization container
        if self.bdp_client and self.config.bdp_connection:
            yield from self._process_organization()

        # Create atomic event dataset (for enrichment lineage)
        yield from self.atomic_event_builder.create_atomic_event_dataset()

        # Extract PII fields from enrichments BEFORE processing schemas
        # This ensures PII tags are available during schema field tagging
        if (
            self.config.field_tagging.enabled
            and self.config.field_tagging.use_pii_enrichment
        ):
            logger.info("Extracting PII fields from enrichment configuration")
            pii_fields = self._extract_pii_fields()
            if pii_fields:
                logger.info(f"Found {len(pii_fields)} PII fields from enrichments")
            else:
                logger.warning(
                    "No PII fields found in enrichment configuration. "
                    "Field tagging will fall back to pattern matching."
                )

        # Extract schemas (via processor)
        if self.schema_processor.is_enabled():
            try:
                yield from self.schema_processor.extract()
            except Exception as e:
                self.error_handler.handle_api_error(
                    error=e,
                    operation="schema extraction",
                    context="schema_processor",
                )

        # Emit tracking plan containers BEFORE event specs so that
        # auto_browse_path_v2 can resolve the full container chain
        # (org → tracking_plan) when computing browse paths for event specs.
        if self.tracking_plan_processor.is_enabled():
            try:
                yield from self.tracking_plan_processor.emit_containers()
            except Exception as e:
                self.error_handler.handle_api_error(
                    error=e,
                    operation="tracking plan container emission",
                    context="tracking_plan_processor",
                )

        # Extract event specifications (via processor)
        if self.event_spec_processor.is_enabled():
            try:
                yield from self.event_spec_processor.extract()
            except Exception as e:
                self.error_handler.handle_api_error(
                    error=e,
                    operation="event specification extraction",
                    context="event_spec_processor",
                )

        # Extract Snowplow standard schemas (via processor) - MUST be after event specs
        # so we have collected all referenced Iglu URIs from event specifications
        if self.standard_schema_processor.is_enabled():
            try:
                yield from self.standard_schema_processor.extract()
            except Exception as e:
                self.error_handler.handle_api_error(
                    error=e,
                    operation="standard schema extraction",
                    context="standard_schema_processor",
                )

        # Emit schema metadata for event specs now that ALL schemas (custom + standard) are extracted
        if self.event_spec_processor.is_enabled():
            try:
                yield from self.event_spec_processor.emit_event_spec_schema_metadata()
            except Exception as e:
                self.error_handler.handle_api_error(
                    error=e,
                    operation="event spec schema metadata emission",
                    context="event_spec_processor",
                )

        # Extract pipelines and enrichments (via processor)
        if self.pipeline_processor.is_enabled():
            try:
                yield from self.pipeline_processor.extract()
            except Exception as e:
                self.error_handler.handle_api_error(
                    error=e,
                    operation="pipeline extraction",
                    context="pipeline_processor",
                )

        # Extract warehouse lineage (via processor)
        if self.warehouse_lineage_processor.is_enabled():
            try:
                yield from self.warehouse_lineage_processor.extract()
            except Exception as e:
                self.error_handler.handle_api_error(
                    error=e,
                    operation="warehouse lineage extraction",
                    context="warehouse_lineage_processor",
                )

        # Compute final statistics
        self.report.compute_stats()

    def _process_organization(self) -> Iterable[MetadataWorkUnit]:
        """
        Emit organization as container.

        Organizations group all schemas and event specifications.
        """
        if not self.config.bdp_connection:
            return

        org_id = self.config.bdp_connection.organization_id

        # Create organization container key
        org_key = SnowplowOrganizationKey(
            organization_id=org_id,
            platform=self.platform,
            instance=self.config.platform_instance,
            env=self.config.env,
        )

        # Use gen_containers to emit all container aspects properly
        yield from gen_containers(
            container_key=org_key,
            name=f"Snowplow Organization ({org_id})",
            sub_types=[DatasetContainerSubTypes.DATABASE],
            description="Snowplow BDP organization containing event and entity schemas",
            extra_properties={
                "organization_id": org_id,
                "platform": "snowplow",
                "connection_mode": self.report.connection_mode,
            },
        )

    def _extract_pii_fields(self) -> set:
        """
        Extract PII fields from PII Pseudonymization enrichment configuration.

        Returns a set of field names that are configured as PII in the
        enrichment config. This is more accurate than pattern matching alone.

        Returns:
            Set of field names that should be classified as PII
        """
        # Return cached value if available
        if self.state.pii_fields_cache is not None:
            return self.state.pii_fields_cache

        pii_fields: set = set()

        # Guard clauses for early exit
        if not self.config.field_tagging.use_pii_enrichment or not self.bdp_client:
            self.state.pii_fields_cache = pii_fields
            return pii_fields

        try:
            pipelines = self.bdp_client.get_pipelines()
            for pipeline in pipelines:
                self._extract_pii_fields_from_pipeline(pipeline.id, pii_fields)
        except Exception as e:
            self.report.report_warning(
                "pii_extraction",
                f"Failed to extract PII fields from enrichments: {e}. "
                "PII field tagging may be incomplete.",
            )
            logger.warning(f"Failed to extract PII fields from enrichments: {e}")

        self.state.pii_fields_cache = pii_fields
        return pii_fields

    def _extract_pii_fields_from_pipeline(
        self, pipeline_id: str, pii_fields: set
    ) -> None:
        """Extract PII fields from a single pipeline's enrichments."""
        assert self.bdp_client is not None  # Caller ensures this
        try:
            enrichments = self.bdp_client.get_enrichments(pipeline_id)
            for enrichment in enrichments:
                self._extract_pii_fields_from_enrichment(enrichment, pii_fields)
        except Exception as e:
            self.report.report_warning(
                "pii_extraction",
                f"Failed to extract PII fields from pipeline {pipeline_id}: {e}. "
                "PII field tagging may be incomplete for this pipeline.",
            )
            logger.warning(
                f"Failed to extract PII fields from pipeline {pipeline_id}: {e}"
            )

    def _extract_pii_fields_from_enrichment(
        self, enrichment: Enrichment, pii_fields: set
    ) -> None:
        """Extract PII fields from a PII Pseudonymization enrichment."""
        # Skip non-PII enrichments
        if not enrichment.schema_ref:
            return
        if "pii_pseudonymization" not in enrichment.schema_ref.lower():
            return

        # Get parameters from enrichment content
        params = self._get_enrichment_parameters(enrichment)
        if not params:
            return

        # Parse using the generic field reference parser
        field_config = FieldReferenceParser.parse_from_key(params, "pii")
        pii_fields.update(field_config.get_all_field_names())

    @staticmethod
    def _get_enrichment_parameters(enrichment: Enrichment) -> Optional[Dict[str, Any]]:
        """Safely extract parameters from enrichment content."""
        if not enrichment.content:
            return None
        if not enrichment.content.data:
            return None
        return enrichment.content.data.parameters

    def get_report(self) -> SnowplowSourceReport:
        return self.report

    def close(self) -> None:
        """Clean up resources after ingestion completes."""
        # Clean up FileBackedDict SQLite resources in IngestionState
        self.state.close()
        if self.bdp_client:
            self.bdp_client.close()
        if self.iglu_client:
            self.iglu_client.close()
        super().close()

    @staticmethod
    def test_connection(config_dict: dict) -> TestConnectionReport:
        """Test connection to Snowplow APIs."""
        test_report = TestConnectionReport()

        # Initialize capability_report if None
        if test_report.capability_report is None:
            test_report.capability_report = {}

        try:
            config = SnowplowSourceConfig.model_validate(config_dict)

            # Test BDP connection
            if config.bdp_connection:
                try:
                    bdp_client = SnowplowBDPClient(config.bdp_connection)
                    if bdp_client.test_connection():
                        test_report.basic_connectivity = CapabilityReport(capable=True)
                        test_report.capability_report[
                            SourceCapability.SCHEMA_METADATA
                        ] = CapabilityReport(capable=True)
                    else:
                        test_report.basic_connectivity = CapabilityReport(
                            capable=False, failure_reason="BDP API connection failed"
                        )
                except Exception as e:
                    test_report.basic_connectivity = CapabilityReport(
                        capable=False, failure_reason=f"BDP connection error: {e}"
                    )

            # Test Iglu connection
            if config.iglu_connection:
                try:
                    iglu_client = IgluClient(config.iglu_connection)
                    if iglu_client.test_connection():
                        test_report.basic_connectivity = CapabilityReport(capable=True)
                        test_report.capability_report[
                            SourceCapability.SCHEMA_METADATA
                        ] = CapabilityReport(capable=True)
                    else:
                        test_report.basic_connectivity = CapabilityReport(
                            capable=False, failure_reason="Iglu connection failed"
                        )
                except Exception as e:
                    test_report.basic_connectivity = CapabilityReport(
                        capable=False, failure_reason=f"Iglu connection error: {e}"
                    )

        except Exception as e:
            test_report.basic_connectivity = CapabilityReport(
                capable=False, failure_reason=f"Configuration error: {e}"
            )

        return test_report



"""Deprecated dataset-only ownership transformers.

The canonical implementations now live in ``add_ownership.py``.  This
module re-exports everything for backward compatibility and provides the
deprecated ``*Dataset*`` wrappers that pin ``entity_types`` to the legacy
set (everything except container).
"""

import warnings
from typing import List

from datahub.errors import DataHubDeprecationWarning
from datahub.ingestion.transformer.add_ownership import (  # noqa: F401 — re-exports
    AddOwnership,
    AddOwnershipConfig as AddDatasetOwnershipConfig,
    OwnershipBaseConfig as DatasetOwnershipBaseConfig,
    PatternAddOwnership,
    PatternOwnershipConfig as PatternDatasetOwnershipConfig,
    SimpleAddOwnership,
    SimpleOwnershipConfig as SimpleDatasetOwnershipConfig,
)

# The legacy *_dataset_* ownership transformers already processed these
# types (everything except container) despite the "dataset" name.
_LEGACY_OWNERSHIP_ENTITY_TYPES: List[str] = [
    "dataset",
    "dataJob",
    "dataFlow",
    "chart",
    "dashboard",
]

_DEPRECATION_MSG = (
    "{cls} is deprecated — use {replacement} instead. "
    "The *Dataset* variants will be removed in a future release."
)


def _emit_deprecation(cls_name: str, replacement: str) -> None:
    warnings.warn(
        _DEPRECATION_MSG.format(cls=cls_name, replacement=replacement),
        DataHubDeprecationWarning,
        stacklevel=3,
    )


class AddDatasetOwnership(AddOwnership):
    """Deprecated: use AddOwnership instead. Legacy wrapper."""

    def __init__(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        _emit_deprecation("AddDatasetOwnership", "AddOwnership")
        super().__init__(*args, **kwargs)

    def entity_types(self) -> List[str]:
        return _LEGACY_OWNERSHIP_ENTITY_TYPES


class SimpleAddDatasetOwnership(SimpleAddOwnership):
    """Deprecated: use SimpleAddOwnership instead. Legacy wrapper."""

    def __init__(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        _emit_deprecation("SimpleAddDatasetOwnership", "SimpleAddOwnership")
        super().__init__(*args, **kwargs)

    def entity_types(self) -> List[str]:
        return _LEGACY_OWNERSHIP_ENTITY_TYPES


class PatternAddDatasetOwnership(PatternAddOwnership):
    """Deprecated: use PatternAddOwnership instead. Legacy wrapper."""

    def __init__(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        _emit_deprecation("PatternAddDatasetOwnership", "PatternAddOwnership")
        super().__init__(*args, **kwargs)

    def entity_types(self) -> List[str]:
        return _LEGACY_OWNERSHIP_ENTITY_TYPES



# Import RestrictedText from your existing module
# Uncomment and adjust the import path as needed:
# from your_restricted_text_module import RestrictedText
# The following is a list of tag constraints:
# You can assign a maximum of 50 tags to a single securable object.
# The maximum length of a tag key is 255 characters.
# The maximum length of a tag value is 1000 characters.
# The following characters are not allowed in tag keys:
# . , - = / :
# Tag search using the workspace search UI is supported only for tables, views, and table columns.
# Tag search requires exact term matching.
# https://learn.microsoft.com/en-us/azure/databricks/database-objects/tags#constraint
from typing import Any, Dict, Optional

from pydantic import field_validator
from typing_extensions import ClassVar

from datahub.api.entities.external.external_tag import ExternalTag
from datahub.api.entities.external.restricted_text import RestrictedText


class LakeFormationTagKeyText(RestrictedText):
    """RestrictedText configured for Lake Formation tag keys."""

    DEFAULT_MAX_LENGTH: ClassVar[int] = 50
    # Lake Formation tag keys restrictions
    DEFAULT_REPLACEMENT_CHAR: ClassVar[str] = "_"
    DEFAULT_TRUNCATION_SUFFIX: ClassVar[str] = ""  # No suffix for clean identifiers


class LakeFormationTagValueText(RestrictedText):
    """RestrictedText configured for Lake Formation tag values."""

    DEFAULT_MAX_LENGTH: ClassVar[int] = 50
    # Lake Formation tag values restrictions
    DEFAULT_REPLACEMENT_CHAR: ClassVar[str] = " "
    DEFAULT_TRUNCATION_SUFFIX: ClassVar[str] = "..."


class LakeFormationTag(ExternalTag):
    """
    A tag type specifically designed for LakeFormation tag restrictions.

    LakeFormation Tag Restrictions:
    - Key: Max 127 characters, alphanumeric + hyphens, underscores, periods only
    - Value: Max 256 characters, more permissive but no control characters
    """

    key: LakeFormationTagKeyText
    value: Optional[LakeFormationTagValueText] = None
    catalog: Optional[str] = None

    @field_validator("key", mode="before")
    @classmethod
    def _validate_key(cls, v: Any) -> LakeFormationTagKeyText:
        """Validate and convert key field."""
        if isinstance(v, LakeFormationTagKeyText):
            return v

        # If we get a RestrictedText object from parent class validation, use its raw_text value
        if hasattr(v, "raw_text"):
            return LakeFormationTagKeyText(raw_text=v.raw_text)

        return LakeFormationTagKeyText(raw_text=v)

    @field_validator("value", mode="before")
    @classmethod
    def _validate_value(cls, v: Any) -> Optional[LakeFormationTagValueText]:
        """Validate and convert value field."""
        if v is None:
            return None

        if isinstance(v, LakeFormationTagValueText):
            return v

        # If we get a RestrictedText object from parent class validation, use its raw_text value
        if hasattr(v, "raw_text"):
            text_value = v.raw_text
            # If value is an empty string, set it to None to not generate empty value in DataHub tag
            if not str(text_value):
                return None
            return LakeFormationTagValueText(raw_text=text_value)

        # If value is an empty string, set it to None to not generate empty value in DataHub tag
        if not str(v):
            return None

        return LakeFormationTagValueText(raw_text=v)

    def __eq__(self, other: object) -> bool:
        """Check equality based on key and value."""
        if not isinstance(other, LakeFormationTag):
            return False
        return str(self.key) == str(other.key) and (
            str(self.value) if self.value else None
        ) == (str(other.value) if other.value else None)

    def __hash__(self) -> int:
        """Make LakeFormationTag hashable based on key and value."""
        return hash((str(self.key), str(self.value) if self.value else None))

    @classmethod
    def from_dict(cls, tag_dict: Dict[str, Any]) -> "LakeFormationTag":
        """
        Create a LakeFormationTag from a dictionary with 'key' and optional 'value'.

        Args:
            tag_dict: Dictionary with 'key' and optional 'value' keys

        Returns:
            LakeFormationTag instance
        """
        return cls(key=tag_dict["key"], value=tag_dict.get("value"))

    @classmethod
    def from_key_value(
        cls, key: str, value: Optional[str] = None
    ) -> "LakeFormationTag":
        """
        Create a LakeFormationTagPlatformResource from explicit key and value.

        Overrides the parent method to return the correct type.

        Args:
            key: Tag key
            value: Optional tag value

        Returns:
            LakeFormationTag instance
        """
        return cls(key=key, value=value)

    def to_dict(self) -> Dict[str, str]:
        """
        Convert to dictionary format suitable for LakeFormation tag.

        Returns:
            Dictionary with 'key' and optionally 'value'
        """
        result: Dict[str, str] = {"key": self.key.raw_text}
        if self.value is not None:
            result["value"] = self.value.raw_text
        return result

    def to_display_dict(self) -> Dict[str, str]:
        """
        Convert to dictionary format showing processed values.

        Returns:
            Dictionary with processed 'key' and optional 'value'
        """
        result: Dict[str, str] = {"key": str(self.key)}
        if self.value is not None:
            result["value"] = str(self.value)
        return result

    def __repr__(self) -> str:
        if self.value:
            return f"LakeFormationTag(key={self.key!r}, value={self.value!r})"
        else:
            return f"LakeFormationTag(key={self.key!r})"



"""Authentication helpers for Microsoft Fabric REST APIs."""

import logging
from typing import Optional

from azure.core.credentials import TokenCredential

from datahub.ingestion.source.azure.azure_auth import AzureCredentialConfig

logger = logging.getLogger(__name__)

# Fabric REST API scope - using Power BI scope as Fabric is part of Power BI ecosystem
# This may need to be updated if a Fabric-specific scope becomes available
FABRIC_API_SCOPE = "https://analysis.windows.net/powerbi/api/.default"

# OneLake Table APIs require Storage audience token
ONELAKE_STORAGE_SCOPE = "https://storage.azure.com/.default"

# SQL Analytics Endpoint requires database scope for Azure AD authentication
SQL_ANALYTICS_ENDPOINT_SCOPE = "https://database.windows.net/.default"


class FabricAuthHelper:
    """Helper class for authenticating with Microsoft Fabric REST APIs.

    Converts Azure TokenCredential to Bearer tokens for REST API calls.
    Handles token refresh and caching.
    """

    def __init__(self, credential_config: AzureCredentialConfig):
        """Initialize the authentication helper.

        Args:
            credential_config: Azure credential configuration
        """
        self.credential_config = credential_config
        self._credential: Optional[TokenCredential] = None
        self._cached_token: Optional[str] = None
        self._cached_scope: Optional[str] = None
        self._token_expires_at: Optional[float] = None

    def get_credential(self) -> TokenCredential:
        """Get the Azure TokenCredential.

        Returns:
            TokenCredential instance
        """
        if self._credential is None:
            self._credential = self.credential_config.get_credential()
        return self._credential

    def get_bearer_token(self, scope: Optional[str] = None) -> str:
        """Get a Bearer token for REST API authentication.

        Args:
            scope: Optional scope/audience. Defaults to FABRIC_API_SCOPE.
                  Use ONELAKE_STORAGE_SCOPE for OneLake Delta Table APIs.
                  Use SQL_ANALYTICS_ENDPOINT_SCOPE for SQL Analytics Endpoint connections.

        Returns:
            Bearer token string (without "Bearer " prefix)

        Raises:
            Exception: If token acquisition fails
        """
        import time

        target_scope = scope or FABRIC_API_SCOPE

        # Check if cached token is still valid and for the same scope
        if (
            self._cached_token is not None
            and self._cached_scope == target_scope
            and self._token_expires_at is not None
            and time.time() < (self._token_expires_at - 300)
        ):
            return self._cached_token

        try:
            credential = self.get_credential()
            token_response = credential.get_token(target_scope)

            self._cached_token = token_response.token
            self._cached_scope = target_scope
            # Token expires_at is a datetime, convert to timestamp
            if hasattr(token_response, "expires_on"):
                # expires_on can be an int (Unix timestamp) or datetime
                expires_on = token_response.expires_on
                if isinstance(expires_on, (int, float)):
                    self._token_expires_at = float(expires_on)
                else:
                    # It's a datetime, convert to timestamp
                    self._token_expires_at = expires_on.timestamp()
            else:
                # Fallback: assume 1 hour validity if expires_on not available
                self._token_expires_at = time.time() + 3600

            logger.debug(f"Successfully acquired token for scope: {target_scope}")
            return self._cached_token

        except Exception as e:
            logger.error(f"Failed to acquire token for scope {target_scope}: {e}")
            raise

    def get_authorization_header(self, scope: Optional[str] = None) -> str:
        """Get the full Authorization header value.

        Args:
            scope: Optional scope/audience. Defaults to FABRIC_API_SCOPE.

        Returns:
            Authorization header value: "Bearer {token}"
        """
        token = self.get_bearer_token(scope=scope)
        return f"Bearer {token}"



import logging
from typing import Dict, List, Optional

import requests

logger = logging.getLogger(__name__)


class KafkaTopicCache:
    """
    Simple topic cache shared across connectors within a single ingestion run.

    Provides cross-connector topic sharing and consumer group assignment caching
    for improved performance during DataHub ingestion runs.
    """

    def __init__(self):
        """Initialize the topic cache."""
        self._topics_cache: Dict[str, List[str]] = {}
        self._consumer_groups_cache: Dict[str, Dict[str, List[str]]] = {}
        self.cache_hits = 0
        self.cache_misses = 0

    def get_all_topics(self, cluster_id: str) -> Optional[List[str]]:
        """
        Get all topics for a cluster from cache.

        Args:
            cluster_id: Unique identifier for the Kafka cluster

        Returns:
            List of topic names if cached, None otherwise
        """
        if cluster_id in self._topics_cache:
            topics = self._topics_cache[cluster_id]
            self.cache_hits += 1
            logger.debug(f"Cache hit for cluster {cluster_id}: {len(topics)} topics")
            return topics.copy()

        self.cache_misses += 1
        return None

    def set_all_topics(self, cluster_id: str, topics: List[str]) -> None:
        """
        Cache topics for a cluster.

        Args:
            cluster_id: Unique identifier for the Kafka cluster
            topics: List of topic names to cache
        """
        self._topics_cache[cluster_id] = topics.copy()
        logger.info(f"Cached {len(topics)} topics for cluster {cluster_id}")

    def get_consumer_group_assignments(
        self, cluster_id: str
    ) -> Optional[Dict[str, List[str]]]:
        """
        Get consumer group assignments from cache.

        Args:
            cluster_id: Unique identifier for the Kafka cluster

        Returns:
            Dict mapping consumer group names to their assigned topics, None if not cached
        """
        if cluster_id in self._consumer_groups_cache:
            assignments = self._consumer_groups_cache[cluster_id]
            self.cache_hits += 1
            logger.debug(
                f"Cache hit for consumer groups in cluster {cluster_id}: {len(assignments)} groups"
            )
            return {group: topics.copy() for group, topics in assignments.items()}

        self.cache_misses += 1
        return None

    def set_consumer_group_assignments(
        self, cluster_id: str, assignments: Dict[str, List[str]]
    ) -> None:
        """
        Cache consumer group assignments for a cluster.

        Args:
            cluster_id: Unique identifier for the Kafka cluster
            assignments: Dict mapping consumer group names to their assigned topics
        """
        cached_assignments = {
            group: topics.copy() for group, topics in assignments.items()
        }
        self._consumer_groups_cache[cluster_id] = cached_assignments
        logger.info(
            f"Cached consumer group assignments for {len(assignments)} groups in cluster {cluster_id}"
        )

    def invalidate_cluster(self, cluster_id: str) -> None:
        """
        Invalidate all cache entries for a specific cluster.

        Args:
            cluster_id: Unique identifier for the Kafka cluster
        """
        removed_topics = self._topics_cache.pop(cluster_id, None)
        removed_groups = self._consumer_groups_cache.pop(cluster_id, None)

        logger.info(
            f"Invalidated cache for cluster {cluster_id}: "
            f"topics={'yes' if removed_topics else 'no'}, "
            f"consumer_groups={'yes' if removed_groups else 'no'}"
        )

    def clear_all(self) -> None:
        """Clear all cached data."""
        topics_count = len(self._topics_cache)
        groups_count = len(self._consumer_groups_cache)

        self._topics_cache.clear()
        self._consumer_groups_cache.clear()
        self.cache_hits = 0
        self.cache_misses = 0

        logger.info(
            f"Cleared all cache data: {topics_count} topic caches, {groups_count} consumer group caches"
        )

    def get_cache_stats(self) -> Dict[str, int]:
        """
        Get cache statistics for reporting.

        Returns:
            Dict with cache hit/miss statistics and entry counts
        """
        return {
            "cache_hits": self.cache_hits,
            "cache_misses": self.cache_misses,
            "topics_cached_clusters": len(self._topics_cache),
            "consumer_groups_cached_clusters": len(self._consumer_groups_cache),
            "hit_rate": int(
                self.cache_hits / max(self.cache_hits + self.cache_misses, 1) * 100
            ),
        }


class ConfluentCloudTopicRetriever:
    """
    Specialized topic retriever for Confluent Cloud environments.

    Integrates with KafkaTopicCache for efficient cross-connector topic sharing.
    """

    def __init__(self, session: requests.Session, cache: KafkaTopicCache):
        """
        Initialize the topic retriever.

        Args:
            session: Configured requests session with authentication
            cache: Shared topic cache instance
        """
        self.session = session
        self.cache = cache

    def get_all_topics_cached(
        self,
        kafka_rest_endpoint: str,
        cluster_id: str,
        auth_headers: Optional[Dict[str, str]] = None,
    ) -> List[str]:
        """
        Get all topics using cache-first strategy.

        Args:
            kafka_rest_endpoint: Kafka REST API v3 endpoint
            cluster_id: Kafka cluster identifier
            auth_headers: Optional authentication headers

        Returns:
            List of topic names, empty list if retrieval fails
        """
        # Try cache first
        cached_topics = self.cache.get_all_topics(cluster_id)
        if cached_topics is not None:
            return cached_topics

        # Cache miss - fetch from API
        try:
            topics = self._fetch_topics_from_api(
                kafka_rest_endpoint, cluster_id, auth_headers
            )
            if topics:
                self.cache.set_all_topics(cluster_id, topics)
                return topics
            else:
                logger.warning(f"No topics retrieved from API for cluster {cluster_id}")
                return []

        except Exception as e:
            logger.warning(
                f"Failed to retrieve topics from API for cluster {cluster_id}: {e}"
            )
            return []

    def _fetch_topics_from_api(
        self,
        kafka_rest_endpoint: str,
        cluster_id: str,
        auth_headers: Optional[Dict[str, str]] = None,
    ) -> List[str]:
        """
        Fetch topics directly from Confluent Cloud Kafka REST API v3.

        Handles pagination by following the metadata.next URL until all topics are retrieved.

        Args:
            kafka_rest_endpoint: Kafka REST API v3 endpoint
            cluster_id: Kafka cluster identifier
            auth_headers: Optional authentication headers

        Returns:
            List of topic names
        """
        # Build initial API URL
        if kafka_rest_endpoint.endswith("/"):
            kafka_rest_endpoint = kafka_rest_endpoint.rstrip("/")
        topics_url = f"{kafka_rest_endpoint}/kafka/v3/clusters/{cluster_id}/topics"

        # Set up headers
        headers = {"Accept": "application/json", "Content-Type": "application/json"}
        if auth_headers:
            headers.update(auth_headers)

        all_topics = []
        current_url = topics_url
        page_count = 0

        # Paginate through all results
        while current_url:
            page_count += 1
            response = self.session.get(current_url, headers=headers)
            response.raise_for_status()

            # Parse v3 API response format
            topics_data = response.json()
            if topics_data.get("kind") == "KafkaTopicList" and "data" in topics_data:
                # Extract non-internal topics from current page
                page_topics = [
                    topic["topic_name"]
                    for topic in topics_data["data"]
                    if not topic.get("is_internal", False)
                ]
                all_topics.extend(page_topics)
                logger.debug(
                    f"Retrieved {len(page_topics)} topics from page {page_count} for cluster {cluster_id}"
                )

                # Check for next page
                metadata = topics_data.get("metadata", {})
                current_url = metadata.get("next")
            else:
                logger.warning(
                    f"Unexpected response format from Kafka REST API for cluster {cluster_id}"
                )
                break

        logger.info(
            f"Retrieved {len(all_topics)} topics across {page_count} page(s) from Confluent Cloud Kafka REST API v3 for cluster {cluster_id}"
        )
        return all_topics

    def get_consumer_group_assignments_cached(
        self,
        kafka_rest_endpoint: str,
        cluster_id: str,
        auth_headers: Optional[Dict[str, str]] = None,
    ) -> Dict[str, List[str]]:
        """
        Get consumer group assignments using cache-first strategy.

        Args:
            kafka_rest_endpoint: Kafka REST API v3 endpoint
            cluster_id: Kafka cluster identifier
            auth_headers: Optional authentication headers

        Returns:
            Dict mapping consumer group names to their assigned topics
        """
        # Try cache first
        cached_assignments = self.cache.get_consumer_group_assignments(cluster_id)
        if cached_assignments is not None:
            return cached_assignments

        # Cache miss - fetch from API
        try:
            assignments = self._fetch_consumer_groups_from_api(
                kafka_rest_endpoint, cluster_id, auth_headers
            )
            if assignments:
                self.cache.set_consumer_group_assignments(cluster_id, assignments)
                return assignments
            else:
                logger.debug(
                    f"No consumer group assignments retrieved for cluster {cluster_id}"
                )
                return {}

        except Exception as e:
            logger.warning(
                f"Failed to retrieve consumer group assignments for cluster {cluster_id}: {e}"
            )
            return {}

    def _fetch_consumer_groups_from_api(
        self,
        kafka_rest_endpoint: str,
        cluster_id: str,
        auth_headers: Optional[Dict[str, str]] = None,
    ) -> Dict[str, List[str]]:
        """
        Fetch consumer group assignments from Confluent Cloud Kafka REST API v3.

        Handles pagination by following the metadata.next URL until all consumer groups are retrieved.

        Args:
            kafka_rest_endpoint: Kafka REST API v3 endpoint
            cluster_id: Kafka cluster identifier
            auth_headers: Optional authentication headers

        Returns:
            Dict mapping consumer group names to their assigned topics
        """
        # Build initial API URL for consumer groups
        if kafka_rest_endpoint.endswith("/"):
            kafka_rest_endpoint = kafka_rest_endpoint.rstrip("/")
        consumer_groups_url = (
            f"{kafka_rest_endpoint}/kafka/v3/clusters/{cluster_id}/consumer-groups"
        )

        # Set up headers
        headers = {"Accept": "application/json", "Content-Type": "application/json"}
        if auth_headers:
            headers.update(auth_headers)

        assignments = {}
        current_url = consumer_groups_url
        page_count = 0

        try:
            # Paginate through all consumer groups
            while current_url:
                page_count += 1
                response = self.session.get(current_url, headers=headers)
                response.raise_for_status()

                groups_data = response.json()
                if (
                    groups_data.get("kind") == "KafkaConsumerGroupList"
                    and "data" in groups_data
                ):
                    for group in groups_data["data"]:
                        group_id = group["consumer_group_id"]

                        # Get assignments for each group
                        assignments_url = f"{consumer_groups_url}/{group_id}/consumers"
                        try:
                            assignments_response = self.session.get(
                                assignments_url, headers=headers
                            )
                            assignments_response.raise_for_status()

                            assignments_data = assignments_response.json()
                            if (
                                assignments_data.get("kind") == "KafkaConsumerList"
                                and "data" in assignments_data
                            ):
                                group_topics = set()
                                for consumer in assignments_data["data"]:
                                    for assignment in consumer.get("assignments", []):
                                        group_topics.add(
                                            assignment.get("topic_name", "")
                                        )

                                if group_topics:
                                    assignments[group_id] = list(group_topics)

                        except Exception as e:
                            logger.debug(
                                f"Failed to get assignments for consumer group {group_id}: {e}"
                            )
                            continue

                    # Check for next page
                    metadata = groups_data.get("metadata", {})
                    current_url = metadata.get("next")
                else:
                    logger.warning(
                        f"Unexpected response format from Kafka REST API for consumer groups in cluster {cluster_id}"
                    )
                    break

            logger.info(
                f"Retrieved assignments for {len(assignments)} consumer groups across {page_count} page(s) in cluster {cluster_id}"
            )

        except Exception as e:
            logger.warning(
                f"Failed to retrieve consumer groups from API for cluster {cluster_id}: {e}"
            )

        return assignments



"""
Data Structure Builder for Snowplow connector.

Handles processing of individual data structures (schemas) including:
- Fetching full schema definitions
- Parsing schema metadata
- Building dataset objects with all aspects
- Field tagging and structured properties
"""

import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
)

from datahub.emitter.mcp_builder import ContainerKey, gen_containers
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.snowplow.builders.container_keys import (
    SnowplowOrganizationKey,
    SnowplowVendorKey,
)
from datahub.ingestion.source.snowplow.constants import (
    DatasetSubtype,
    EventFieldType,
    SchemaType,
    get_schema_subtype,
)
from datahub.ingestion.source.snowplow.models.snowplow_models import (
    DataStructure,
    DataStructureDeployment,
    SchemaMetadata as SnowplowSchemaMetadata,
)
from datahub.ingestion.source.snowplow.utils.schema_parser import SnowplowSchemaParser
from datahub.metadata.schema_classes import (
    ContainerClass,
    GlobalTagsClass,
    OwnerClass,
    SchemaFieldClass,
    SchemaMetadataClass,
    StatusClass,
    TagAssociationClass,
)
from datahub.sdk.dataset import Dataset

if TYPE_CHECKING:
    from datahub.ingestion.source.snowplow.dependencies import (
        IngestionState,
        ProcessorDependencies,
    )
    from datahub.ingestion.source.snowplow.services.column_lineage_builder import (
        ColumnLineageBuilder,
    )
    from datahub.ingestion.source.snowplow.services.field_tagging import FieldTagContext

logger = logging.getLogger(__name__)


class DataStructureBasicInfo(NamedTuple):
    """Validated basic information extracted from a DataStructure."""

    vendor: str
    name: str
    schema_meta: SnowplowSchemaMetadata
    schema_type: str
    version: str
    schema_identifier: str


class DataStructureBuilder:
    """
    Builder for processing Snowplow data structures (schemas).

    Handles the complete lifecycle of processing a data structure from the BDP API,
    including fetching full definitions, parsing schemas, building datasets, and
    emitting all related metadata.
    """

    def __init__(
        self,
        deps: "ProcessorDependencies",
        state: "IngestionState",
        column_lineage_builder: "ColumnLineageBuilder",
    ):
        """
        Initialize data structure builder.

        Args:
            deps: Explicit dependencies (clients, builders, config, etc.)
            state: Shared mutable state
            column_lineage_builder: Builder for column-level lineage
        """
        self.deps = deps
        self.state = state
        self.column_lineage_builder = column_lineage_builder

        # Convenience accessors
        self.config = deps.config
        self.report = deps.report
        self.bdp_client = deps.bdp_client
        self.platform = deps.platform
        self.urn_factory = deps.urn_factory
        self.ownership_builder = deps.ownership_builder
        self.field_tagger = deps.field_tagger

        # Track emitted vendor containers to avoid duplicates
        self._emitted_vendors: Set[str] = set()

    def process_data_structure(
        self, data_structure: DataStructure
    ) -> Iterable[MetadataWorkUnit]:
        """
        Process a single data structure (schema).

        Args:
            data_structure: Data structure from BDP API

        Yields:
            MetadataWorkUnit: Schema metadata work units
        """
        # Fetch full schema definition if missing
        data_structure = self._fetch_full_schema_definition(data_structure)

        # Validate and extract basic info
        basic_info = self._validate_and_extract_basic_info(data_structure)
        if not basic_info:
            return

        vendor, name, schema_meta, schema_type, version, schema_identifier = basic_info

        # Track found schema
        self.report.report_schema_found(schema_type)

        # Apply filtering
        if not self._should_process_schema(schema_identifier, schema_type, schema_meta):
            return

        # Capture first event schema for parsed events dataset naming
        self._capture_first_event_schema(vendor, name, schema_type)

        # Emit vendor container (idempotent - only emits once per vendor)
        yield from self._emit_vendor_container(vendor)

        # Build dataset name
        dataset_name = self._build_dataset_name(vendor, name, version)

        # Build dataset properties
        custom_properties = self._build_custom_properties(
            data_structure, vendor, name, version, schema_type, schema_meta
        )

        # Build owners
        owners_list = self._build_owners_list(data_structure, schema_identifier)

        # Build subtype and extra aspects
        subtype = (
            get_schema_subtype(schema_type) if schema_type else DatasetSubtype.SCHEMA
        )
        extra_aspects = self._build_extra_aspects(name, schema_type)

        # Set container via extra_aspects; browse path is computed by auto_browse_path_v2
        leaf_container_urn = self._get_leaf_container_urn(vendor)
        if leaf_container_urn:
            extra_aspects.append(ContainerClass(container=leaf_container_urn))

        # Parse schema metadata and add tags
        schema_metadata = self._parse_schema_metadata(
            data_structure,
            vendor,
            name,
            version,
            schema_type,
            schema_identifier,
            extra_aspects,
        )

        # Create and yield Dataset
        dataset = self._create_dataset(
            dataset_name,
            name,
            data_structure,
            version,
            custom_properties,
            subtype,
            owners_list,
            extra_aspects,
        )
        yield from dataset.as_workunits()

        # Post-processing: emit structured properties, cache results, emit lineage
        yield from self._post_process_dataset(
            dataset, data_structure, schema_metadata, vendor, name, version, schema_type
        )

        # Track extracted schema
        self.report.report_schema_extracted(schema_type)

    def _validate_and_extract_basic_info(
        self, data_structure: DataStructure
    ) -> Optional[DataStructureBasicInfo]:
        """Validate and extract basic information from data structure."""
        if (
            not data_structure.vendor
            or not data_structure.name
            or not data_structure.meta
        ):
            logger.warning(
                f"Data structure missing basic fields (vendor, name, or meta), skipping: {data_structure.hash}"
            )
            return None

        vendor = data_structure.vendor
        name = data_structure.name
        schema_meta = data_structure.meta
        schema_type = schema_meta.schema_type or "event"

        version = self._get_schema_version(data_structure)
        if not version:
            logger.warning(f"Cannot determine version for {vendor}/{name}, skipping")
            return None

        schema_identifier = f"{vendor}/{name}"
        return DataStructureBasicInfo(
            vendor=vendor,
            name=name,
            schema_meta=schema_meta,
            schema_type=schema_type,
            version=version,
            schema_identifier=schema_identifier,
        )

    def _should_process_schema(
        self,
        schema_identifier: str,
        schema_type: str,
        schema_meta: SnowplowSchemaMetadata,
    ) -> bool:
        """Check if schema should be processed based on filters."""
        if not self.config.schema_pattern.allowed(schema_identifier):
            self.report.report_schema_filtered(schema_type, schema_identifier)
            return False

        if schema_type not in self.config.schema_types_to_extract:
            logger.debug(f"Skipping schema {schema_identifier} with type {schema_type}")
            return False

        if schema_meta.hidden and not self.config.include_hidden_schemas:
            self.report.report_hidden_schema(skipped=True)
            logger.debug(f"Skipping hidden schema: {schema_identifier}")
            return False
        elif schema_meta.hidden:
            self.report.report_hidden_schema(skipped=False)

        return True

    def _capture_first_event_schema(
        self, vendor: str, name: str, schema_type: str
    ) -> None:
        """Capture first event schema for parsed events dataset naming."""
        if (
            schema_type == SchemaType.EVENT.value
            and self.state.first_event_schema_vendor is None
        ):
            self.state.first_event_schema_vendor = vendor
            self.state.first_event_schema_name = name
            logger.debug(
                f"Captured first event schema for Event dataset naming: {vendor}/{name}"
            )

    def _build_dataset_name(self, vendor: str, name: str, version: str) -> str:
        """Build dataset name based on configuration."""
        if self.config.include_version_in_urn:
            return f"{vendor}.{name}.{version}".replace("/", ".")
        return f"{vendor}.{name}".replace("/", ".")

    def _build_custom_properties(
        self,
        data_structure: DataStructure,
        vendor: str,
        name: str,
        version: str,
        schema_type: str,
        schema_meta: SnowplowSchemaMetadata,
    ) -> Dict[str, str]:
        """Build custom properties for dataset."""
        custom_properties = {
            "vendor": vendor,
            "schemaVersion": version,
            "schema_type": schema_type or "unknown",
            "hidden": str(schema_meta.hidden),
            "igluUri": f"iglu:{vendor}/{name}/jsonschema/{version}",
            **schema_meta.custom_data,
        }
        if data_structure.data:
            custom_properties["format"] = data_structure.data.self_descriptor.format
        return custom_properties

    def _get_leaf_container_urn(self, vendor: str) -> Optional[str]:
        """Return the URN of the leaf vendor container for this vendor namespace.

        Only ContainerClass is set on the dataset; browsePathsV2 is computed
        automatically by the framework's auto_browse_path_v2 stream processor
        from the container hierarchy emitted by _emit_vendor_container.
        """
        if not self.config.bdp_connection:
            return None

        org_id = self.config.bdp_connection.organization_id
        vendor_key = SnowplowVendorKey(
            organization_id=org_id,
            vendor=vendor,
            platform=self.platform,
            instance=self.config.platform_instance,
            env=self.config.env,
        )
        return str(vendor_key.as_urn())

    def _emit_vendor_container(self, vendor: str) -> Iterable[MetadataWorkUnit]:
        """Emit nested vendor containers by splitting the vendor namespace on dots.

        For example, 'com.snowplowanalytics.snowplow' emits three containers:
          Organization → com → snowplowanalytics → snowplow
        Shared prefixes (e.g., 'com') are emitted once and reused.
        """
        if not self.config.bdp_connection:
            return

        org_id = self.config.bdp_connection.organization_id
        segments = vendor.split(".")

        for i in range(len(segments)):
            vendor_path = ".".join(segments[: i + 1])

            if vendor_path in self._emitted_vendors:
                continue
            self._emitted_vendors.add(vendor_path)

            if i == 0:
                parent_key: ContainerKey = SnowplowOrganizationKey(
                    organization_id=org_id,
                    platform=self.platform,
                    instance=self.config.platform_instance,
                    env=self.config.env,
                )
            else:
                parent_path = ".".join(segments[:i])
                parent_key = SnowplowVendorKey(
                    organization_id=org_id,
                    vendor=parent_path,
                    platform=self.platform,
                    instance=self.config.platform_instance,
                    env=self.config.env,
                )

            vendor_key = SnowplowVendorKey(
                organization_id=org_id,
                vendor=vendor_path,
                platform=self.platform,
                instance=self.config.platform_instance,
                env=self.config.env,
            )

            yield from gen_containers(
                container_key=vendor_key,
                name=segments[i],
                sub_types=[DatasetSubtype.SCHEMA],
                parent_container_key=parent_key,
                extra_properties={
                    "vendor": vendor_path,
                },
            )

    def _build_owners_list(
        self, data_structure: DataStructure, schema_identifier: str
    ) -> Optional[List[OwnerClass]]:
        """Build owners list from deployments."""
        if data_structure.deployments and self.bdp_client:
            created_by, modified_by = (
                self.ownership_builder.extract_ownership_from_deployments(
                    data_structure.deployments
                )
            )
            return self.ownership_builder.build_ownership_list(
                created_by, modified_by, schema_identifier
            )
        return None

    def _build_extra_aspects(self, name: str, schema_type: str) -> List[Any]:
        """Build extra aspects including status and optional tags."""
        extra_aspects: List[Any] = [StatusClass(removed=False)]

        logger.debug(
            f"Dataset tag check for '{name}': field_tagging.enabled={self.config.field_tagging.enabled}, "
            f"tag_event_type={self.config.field_tagging.tag_event_type}"
        )

        if (
            self.config.field_tagging.enabled
            and self.config.field_tagging.tag_event_type
        ):
            logger.debug(f"Building dataset tags for '{name}'")
            dataset_tags = self._build_dataset_tags(name, schema_type)
            if dataset_tags:
                extra_aspects.append(dataset_tags)
                logger.debug(f"Added dataset tags to '{name}'")
        else:
            logger.debug(
                f"Skipping dataset tags for '{name}' (tag_event_type={self.config.field_tagging.tag_event_type})"
            )

        return extra_aspects

    def _parse_schema_metadata(
        self,
        data_structure: DataStructure,
        vendor: str,
        name: str,
        version: str,
        schema_type: str,
        schema_identifier: str,
        extra_aspects: List[Any],
    ) -> Optional[SchemaMetadataClass]:
        """Parse schema metadata and add field tags if enabled."""
        if not data_structure.data:
            logger.info(
                f"Schema definition not available for {vendor}/{name}/{version}, skipping detailed schema metadata"
            )
            return None

        try:
            schema_metadata = SnowplowSchemaParser.parse_schema(
                schema_data=data_structure.data.model_dump(),
                vendor=vendor,
                name=name,
                version=version,
            )

            if self.config.field_tagging.enabled:
                schema_metadata = self._add_field_tags(
                    schema_metadata=schema_metadata,
                    data_structure=data_structure,
                    version=version,
                    schema_type=schema_type,
                )

            extra_aspects.append(schema_metadata)
            return schema_metadata

        except Exception as e:
            error_msg = f"Failed to parse schema {schema_identifier}: {e}"
            self.report.report_schema_parsing_error(error_msg)
            logger.error(error_msg)
            return None

    def _create_dataset(
        self,
        dataset_name: str,
        name: str,
        data_structure: DataStructure,
        version: str,
        custom_properties: Dict[str, str],
        subtype: str,
        owners_list: Optional[List[OwnerClass]],
        extra_aspects: List[Any],
    ) -> Dataset:
        """Create Dataset object using SDK V2.

        ContainerClass is set via extra_aspects; browsePathsV2 is left unset
        so auto_browse_path_v2 can compute it from the container hierarchy.
        """
        return Dataset(
            platform=self.platform,
            name=dataset_name,
            platform_instance=self.config.platform_instance,
            env=self.config.env,
            description=data_structure.data.description
            if data_structure.data
            else None,
            display_name=name,
            external_url=self._get_schema_url(
                data_structure.vendor or "",
                name,
                version,
                data_structure.hash,
            ),
            custom_properties=custom_properties,
            subtype=subtype,
            owners=owners_list,
            extra_aspects=extra_aspects,
        )

    def _post_process_dataset(
        self,
        dataset: Dataset,
        data_structure: DataStructure,
        schema_metadata: Optional[SchemaMetadataClass],
        vendor: str,
        name: str,
        version: str,
        schema_type: str,
    ) -> Iterable[MetadataWorkUnit]:
        """Post-process dataset: emit structured properties, cache results, emit lineage."""
        # Emit field-level structured properties if enabled
        if (
            self.config.field_tagging.enabled
            and self.config.field_tagging.use_structured_properties
            and schema_metadata
        ):
            yield from self._emit_field_structured_properties(
                dataset_urn=str(dataset.urn),
                schema_metadata=schema_metadata,
                data_structure=data_structure,
                version=version,
                schema_type=schema_type,
            )

        # Cache the schema URN for pipeline collector job
        self.state.extracted_schema_urns.append(str(dataset.urn))

        # Cache schema fields for Event dataset (indexed by URN for O(1) lookups)
        if schema_metadata and schema_metadata.fields:
            self.state.register_schema_fields(str(dataset.urn), schema_metadata.fields)

        # Emit column-level lineage
        if schema_metadata:
            yield from self.column_lineage_builder.emit_column_lineage(
                dataset_urn=str(dataset.urn),
                vendor=vendor,
                name=name,
                version=version,
                schema_metadata=schema_metadata,
            )

    def _fetch_full_schema_definition(
        self, data_structure: DataStructure
    ) -> DataStructure:
        """
        Fetch full schema definition if only minimal info available.

        Args:
            data_structure: Data structure from BDP API

        Returns:
            Updated data structure with full schema definition (if available)
        """
        if (
            data_structure.data is not None
            or not data_structure.hash
            or not self.bdp_client
        ):
            return data_structure

        # Use version from most recent PROD deployment, fall back to any deployment
        latest_deployment = data_structure.get_latest_deployment(prefer_env="PROD")
        if not latest_deployment:
            logger.warning(
                f"No deployments found for {data_structure.vendor}/{data_structure.name}, cannot determine version to fetch"
            )
            return data_structure

        version = latest_deployment.version
        env = latest_deployment.env

        logger.info(
            f"Fetching schema definition for {data_structure.vendor}/{data_structure.name} version {version} from {env}"
        )

        # Fetch using the /versions/{version} endpoint which returns full schema
        try:
            full_structure = self.bdp_client.get_data_structure_version(
                data_structure.hash, version, env
            )
        except Exception as e:
            logger.warning(
                f"Failed to fetch schema definition for {data_structure.vendor}/{data_structure.name} "
                f"version {version}: {e}. Schema will be extracted without field-level metadata.",
                exc_info=True,
            )
            return data_structure

        if full_structure and full_structure.data:
            logger.info(
                f"Successfully fetched schema definition with {len(full_structure.data.properties or {})} properties"
            )
            # Preserve metadata from list response
            full_structure.meta = data_structure.meta
            full_structure.deployments = data_structure.deployments
            return full_structure
        else:
            logger.warning(
                f"Could not fetch schema definition for {data_structure.hash}/{version}, skipping field-level metadata"
            )
            return data_structure

    def _get_schema_version(self, data_structure: DataStructure) -> Optional[str]:
        """
        Extract schema version from data structure.

        Args:
            data_structure: Data structure from BDP API

        Returns:
            Schema version string, or None if cannot be determined
        """
        # Get version from schema definition if available
        if data_structure.data:
            return data_structure.data.self_descriptor.version

        # Otherwise use version from latest deployment
        if data_structure.deployments:
            sorted_deployments = sorted(
                data_structure.deployments, key=lambda d: d.ts or "", reverse=True
            )
            version = sorted_deployments[0].version
            logger.info(
                f"Schema definition missing, using version from latest deployment: {version}"
            )
            return version

        return None

    def _build_dataset_tags(
        self, name: str, schema_type: Optional[str]
    ) -> Optional[GlobalTagsClass]:
        """
        Build dataset-level tags.

        Dataset-level tags apply to the schema/dataset as a whole, not individual fields.
        Currently adds:
        - Event type tag: Category derived from schema name (e.g., "checkout" from "checkout_started")

        Args:
            name: Schema name (e.g., "checkout_started")
            schema_type: Schema type ("event" or "entity")

        Returns:
            GlobalTagsClass with dataset-level tags, or None if no tags to add
        """
        tags = set()

        # Event type tag (category derived from name)
        # Extract first word: checkout_started → checkout
        event_name = name.split("_")[0] if "_" in name else name
        event_type_tag = self.config.field_tagging.event_type_pattern.format(
            name=event_name
        )
        tags.add(event_type_tag)

        # Convert to TagAssociationClass
        tag_associations = [
            TagAssociationClass(tag=f"urn:li:tag:{tag}") for tag in sorted(tags)
        ]

        return GlobalTagsClass(tags=tag_associations)

    def _add_field_tags(
        self,
        schema_metadata: SchemaMetadataClass,
        data_structure: DataStructure,
        version: str,
        schema_type: str,
    ) -> SchemaMetadataClass:
        """
        Add field-level tags to schema metadata.

        Field tags provide semantic meaning about what kind of data each field contains:
        - PII: Personal identifiable information
        - Email, Phone, IP Address, etc.
        - Event metadata (timestamp, ID, etc.)

        Args:
            schema_metadata: Parsed schema metadata with fields
            data_structure: Data structure from BDP API
            version: Schema version
            schema_type: Schema type ("event" or "entity")

        Returns:
            Updated schema metadata with field tags
        """
        if not schema_metadata.fields:
            return schema_metadata

        pii_fields = self._get_pii_fields_if_enabled()
        event_type = self._get_event_type(schema_type)
        schema_key = f"{data_structure.vendor}/{data_structure.name}"
        field_version_map = self._get_field_version_map(schema_key)

        # Update field descriptions with version information
        self._update_field_descriptions_with_versions(
            schema_metadata.fields, schema_key, field_version_map
        )

        # Add tags to each field
        for field in schema_metadata.fields:
            context = self._build_field_context(
                field=field,
                data_structure=data_structure,
                version=version,
                field_version_map=field_version_map,
                pii_fields=pii_fields,
                event_type=event_type,
            )
            field_tags = self.field_tagger.generate_tags(context)
            if field_tags:
                field.globalTags = field_tags

        return schema_metadata

    def _get_pii_fields_if_enabled(self) -> set:
        """Get PII fields if PII enrichment is enabled."""
        if self.config.field_tagging.use_pii_enrichment:
            return self._get_pii_fields()
        return set()

    def _get_event_type(self, schema_type: str) -> str:
        """Map schema_type to event_type for field tagging."""
        return (
            EventFieldType.SELF_DESCRIBING.value
            if schema_type == SchemaType.EVENT.value
            else EventFieldType.CONTEXT.value
        )

    def _get_field_version_map(self, schema_key: str) -> Dict[str, str]:
        """Get field version map if version tracking is enabled."""
        if (
            self.config.field_tagging.track_field_versions
            and self.config.field_tagging.tag_schema_version
        ):
            return self.state.field_version_cache.get(schema_key, {})
        return {}

    def _update_field_descriptions_with_versions(
        self,
        fields: List[SchemaFieldClass],
        schema_key: str,
        field_version_map: Dict[str, str],
    ) -> None:
        """Update field descriptions with version information."""
        if not (
            self.config.field_tagging.track_field_versions
            and self.config.field_tagging.tag_schema_version
        ):
            return

        if not field_version_map:
            logger.warning(
                f"Field version map is empty for {schema_key} - "
                "field version info will not be added"
            )
            return

        logger.debug(
            f"Applying field version info to {len(fields)} fields in {schema_key} "
            f"(field_version_map has {len(field_version_map)} entries)"
        )

        for field in fields:
            field_added_version = field_version_map.get(field.fieldPath)
            if not field_added_version or field_added_version == "1-0-0":
                continue

            version_suffix = f" (Added in version {field_added_version})"
            if field.description and version_suffix not in field.description:
                field.description = f"{field.description}{version_suffix}"
            elif not field.description:
                field.description = f"Added in version {field_added_version}"

    def _get_field_deployment_info(
        self, deployments: Optional[List[DataStructureDeployment]], field_version: str
    ) -> Tuple[Optional[str], Optional[str]]:
        """
        Get deployment info for a specific field version.

        Args:
            deployments: List of deployments from data structure
            field_version: The version when the field was added

        Returns:
            Tuple of (initiator, timestamp), both may be None
        """
        if not deployments:
            return None, None

        matching_deployments = [d for d in deployments if d.version == field_version]
        if not matching_deployments:
            return None, None

        deployment = sorted(
            matching_deployments, key=lambda d: d.ts or "", reverse=True
        )[0]
        return deployment.initiator, deployment.ts

    def _build_field_context(
        self,
        field: SchemaFieldClass,
        data_structure: DataStructure,
        version: str,
        field_version_map: Dict[str, str],
        pii_fields: set,
        event_type: str,
    ) -> "FieldTagContext":
        """
        Build field tag context for a single field.

        Args:
            field: Schema field to build context for
            data_structure: Data structure containing field
            version: Schema version (fallback if field version unknown)
            field_version_map: Map of field paths to introduction versions
            pii_fields: Set of PII field paths
            event_type: Event type (self_describing or context)

        Returns:
            FieldTagContext for the field
        """
        from datahub.ingestion.source.snowplow.services.field_tagging import (
            FieldTagContext,
        )

        field_version = field_version_map.get(field.fieldPath, version)
        initiator, timestamp = self._get_field_deployment_info(
            data_structure.deployments, field_version
        )

        return FieldTagContext(
            schema_version=field_version,
            vendor=data_structure.vendor or "",
            name=data_structure.name or "",
            field_name=field.fieldPath,
            field_type=field.nativeDataType,
            field_description=field.description,
            deployment_initiator=initiator,
            deployment_timestamp=timestamp,
            pii_fields=pii_fields,
            event_type=event_type,
            skip_version_tag=field_version == "1-0-0",
        )

    def _get_pii_fields(self) -> set:
        """
        Get PII fields from cache.

        PII fields are extracted from enrichment configuration by the main source
        and cached in the state. This method simply returns the cached value.

        Returns:
            Set of field paths marked as PII (empty set if not available)
        """
        # Return cached value if available
        if self.state.pii_fields_cache is not None:
            return self.state.pii_fields_cache

        # PII fields haven't been extracted yet - return empty set
        # The main source will extract them from enrichments if enabled
        return set()

    def _emit_field_structured_properties(
        self,
        dataset_urn: str,
        schema_metadata: SchemaMetadataClass,
        data_structure: DataStructure,
        version: str,
        schema_type: str,
    ) -> Iterable[MetadataWorkUnit]:
        """
        Emit field-level structured properties.

        Structured properties provide strongly-typed, searchable metadata on fields.
        Currently supports:
        - PII indicators
        - Semantic types (email, phone, IP, etc.)

        Args:
            dataset_urn: URN of the dataset
            schema_metadata: Parsed schema metadata
            data_structure: Data structure from BDP API
            version: Schema version

        Yields:
            MetadataWorkUnit: Field structured property work units
        """
        if not schema_metadata.fields:
            return

        pii_fields = self._get_pii_fields_if_enabled()
        event_type = self._get_event_type(schema_type)
        schema_key = f"{data_structure.vendor}/{data_structure.name}"
        field_version_map = self._get_field_version_map(schema_key)

        if field_version_map:
            logger.debug(
                f"Using field version map for structured properties in {schema_key}"
            )

        for field in schema_metadata.fields:
            context = self._build_field_context(
                field=field,
                data_structure=data_structure,
                version=version,
                field_version_map=field_version_map,
                pii_fields=pii_fields,
                event_type=event_type,
            )

            yield from self.field_tagger.generate_field_structured_properties(
                dataset_urn=dataset_urn,
                field=field,
                context=context,
                field_properties_cache=self.state.field_structured_properties,
            )

    def _get_schema_url(
        self,
        vendor: str,
        name: str,
        version: str,
        schema_hash: Optional[str],
    ) -> Optional[str]:
        """
        Build external URL for schema in BDP Console.

        Args:
            vendor: Schema vendor
            name: Schema name
            version: Schema version
            schema_hash: Schema hash (if available)

        Returns:
            URL to schema in BDP Console, or None if not using BDP
        """
        if not self.config.bdp_connection or not schema_hash:
            return None

        # BDP Console schema URL format:
        # https://console.snowplowanalytics.com/{{org_id}}/data-structures/{{hash}}?version={{version}}
        org_id = self.config.bdp_connection.organization_id
        return f"https://console.snowplowanalytics.com/{org_id}/data-structures/{schema_hash}?version={version}"



"""Google Dataplex source for DataHub ingestion.

This source extracts metadata from Google Dataplex Universal Catalog, including:
- Entries (Universal Catalog) as Datasets with source platform URNs (bigquery, gcs, etc.)
- BigQuery Projects as Containers (project-level containers)
- BigQuery Datasets as Containers (dataset-level containers, nested under project containers)

Reference implementation based on VertexAI and BigQuery V2 sources.
"""

import logging
from itertools import product
from typing import Iterable, Optional

from google.api_core import exceptions
from google.cloud import dataplex_v1
from google.cloud.datacatalog_lineage import LineageClient
from google.oauth2 import service_account

from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import (
    CapabilityReport,
    MetadataWorkUnitProcessor,
    SourceCapability,
    TestableSource,
    TestConnectionReport,
)
from datahub.ingestion.api.source_helpers import auto_workunit
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.dataplex.dataplex_config import DataplexConfig
from datahub.ingestion.source.dataplex.dataplex_entries import (
    DataplexEntriesProcessor,
)
from datahub.ingestion.source.dataplex.dataplex_helpers import EntryDataTuple
from datahub.ingestion.source.dataplex.dataplex_lineage import DataplexLineageExtractor
from datahub.ingestion.source.dataplex.dataplex_report import DataplexReport
from datahub.ingestion.source.state.redundant_run_skip_handler import (
    RedundantLineageRunSkipHandler,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionSourceBase,
)

logger = logging.getLogger(__name__)


@platform_name("Dataplex", id="dataplex")
@config_class(DataplexConfig)
@support_status(SupportStatus.INCUBATING)
@capability(
    SourceCapability.CONTAINERS,
    "Enabled by default",
)
@capability(
    SourceCapability.SCHEMA_METADATA,
    "Enabled by default, can be disabled via configuration `include_schema`",
)
@capability(
    SourceCapability.DESCRIPTIONS,
    "Enabled by default",
)
@capability(
    SourceCapability.LINEAGE_COARSE,
    "Optionally enabled via configuration `include_lineage`",
)
@capability(
    SourceCapability.DELETION_DETECTION,
    "Enabled by default via stateful ingestion",
)
@capability(
    SourceCapability.PLATFORM_INSTANCE,
    "Project containers are generated instead",
    supported=False,
)
@capability(
    SourceCapability.TEST_CONNECTION,
    "Enabled by default",
)
class DataplexSource(StatefulIngestionSourceBase, TestableSource):
    """Source to ingest metadata from Google Dataplex Universal Catalog.

    To add support for a new Dataplex entry type, first define its identity model
    in ``dataplex_ids.py`` by adding a SchemaKey class at the correct level in the
    existing hierarchy (for example project -> instance -> database -> table). Then
    register the entry type in ``DATAPLEX_ENTRY_TYPE_MAPPINGS`` with required fields:
    DataHub platform, DataHub entity type (Container or Dataset), subtype constant
    from ``subtypes.py``, FQN regex, parent-entry regex (if applicable), and key classes.
    FQN formats and parent hierarchy must be derived from:
    https://cloud.google.com/dataplex/docs/fully-qualified-names

    Decide whether the entry maps to a Container or Dataset based on whether it is
    a logical grouping level (instance/database/dataset) versus a concrete asset
    (table/topic). Use subtype constants from ``datahub.ingestion.source.common.subtypes``.
    Top-level containers may not have ``parent_entry`` and should rely on key hierarchy
    for parent traversal. Also note that Spanner entries are currently fetched through
    a search workaround, which bypasses entry-group filtering and must be controlled
    with entry-level filters.

    Lineage extraction strategy:
    - The source first traverses all entries, then performs lineage lookups for each
      entry across the full configured ``(project_id, location)`` scan matrix.
    - This is required because the Dataplex Lineage API does not provide a bulk
      lineage retrieval endpoint that can return all lineage for a project in one call.
    - For graph construction, upstream lookup is sufficient: we query links with
      ``target=<entry_fqn>`` and build DataHub upstream edges from returned sources.
    - The tradeoff is API call volume: every entry is queried separately against every
      configured project-location pair.
    """

    platform: str = "dataplex"

    def _resolve_lineage_project_location_pairs(self) -> list[tuple[str, str]]:
        """Resolve and report lineage scan pairs from configured project/location product."""
        lineage_project_location_pairs = list(
            product(self.config.project_ids, self.config.lineage_locations)
        )
        self.report.info(
            title="Lineage extraction project/location pairs",
            message=(
                "Extracting lineage for configured project/location pairs. "
                "Lineage scan scope is derived from the Cartesian product of "
                "project_ids and lineage_locations."
            ),
            context=str(
                dict(
                    lineage_locations=self.config.lineage_locations,
                    lineage_project_location_pairs=lineage_project_location_pairs,
                )
            ),
        )
        return lineage_project_location_pairs

    def __init__(self, ctx: PipelineContext, config: DataplexConfig):
        super().__init__(config, ctx)
        self.config = config
        self.report: DataplexReport = DataplexReport()

        # Track entry data for lineage extraction across all configured projects.
        self.entry_data: list[EntryDataTuple] = []

        creds = self.config.get_credentials()
        credentials = (
            service_account.Credentials.from_service_account_info(creds)
            if creds
            else None
        )

        # Catalog client for Entry Groups and Entries extraction
        self.catalog_client = dataplex_v1.CatalogServiceClient(credentials=credentials)

        # Initialize redundant lineage run skip handler for stateful lineage ingestion.
        # TODO: Wire this into DataplexLineageExtractor execution flow so lineage API calls
        # can be short-circuited for redundant runs.
        redundant_lineage_run_skip_handler: Optional[RedundantLineageRunSkipHandler] = (
            None
        )
        if self.config.enable_stateful_lineage_ingestion:
            redundant_lineage_run_skip_handler = RedundantLineageRunSkipHandler(
                source=self,
                config=self.config,
                pipeline_name=self.ctx.pipeline_name,
                run_id=self.ctx.run_id,
            )

        if self.config.include_lineage:
            self.lineage_client: Optional[LineageClient] = LineageClient(
                credentials=credentials
            )
            self.lineage_extractor: Optional[DataplexLineageExtractor] = (
                DataplexLineageExtractor(
                    config=self.config,
                    report=self.report.lineage_report,
                    source_report=self.report,
                    lineage_client=self.lineage_client,
                    redundant_run_skip_handler=redundant_lineage_run_skip_handler,
                )
            )
        else:
            self.lineage_client = None
            self.lineage_extractor = None

        self.entries_processor = DataplexEntriesProcessor(
            config=self.config,
            catalog_client=self.catalog_client,
            report=self.report.entries_report,
            entry_data=self.entry_data,
            source_report=self.report,
        )

    @staticmethod
    def test_connection(config_dict: dict) -> TestConnectionReport:
        """Test connection to Dataplex API."""
        test_report = TestConnectionReport()
        try:
            config = DataplexConfig.model_validate(config_dict)
            creds = config.get_credentials()
            credentials = (
                service_account.Credentials.from_service_account_info(creds)
                if creds
                else None
            )

            # Test connection by attempting to list entry groups
            catalog_client = dataplex_v1.CatalogServiceClient(credentials=credentials)
            if config.project_ids:
                project_id = config.project_ids[0]
                parent = (
                    f"projects/{project_id}/locations/{config.entries_locations[0]}"
                )
                entry_groups_request = dataplex_v1.ListEntryGroupsRequest(parent=parent)
                # Just iterate once to verify access
                for _ in catalog_client.list_entry_groups(request=entry_groups_request):
                    break

            test_report.basic_connectivity = CapabilityReport(capable=True)
        except exceptions.GoogleAPICallError as e:
            test_report.basic_connectivity = CapabilityReport(
                capable=False, failure_reason=f"Failed to connect to Dataplex: {e}"
            )
        except Exception as e:
            test_report.basic_connectivity = CapabilityReport(
                capable=False, failure_reason=f"Unexpected error: {e}"
            )

        return test_report

    def get_report(self) -> DataplexReport:
        """Return the ingestion report."""
        return self.report

    def get_workunit_processors(self) -> list[Optional[MetadataWorkUnitProcessor]]:
        """
        Get workunit processors for stateful ingestion.

        Returns processors for:
        - Stale entity removal (deletion detection)
        """
        return [
            *super().get_workunit_processors(),
            StaleEntityRemovalHandler.create(
                self, self.config, self.ctx
            ).workunit_processor,
        ]

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        """Main function to fetch and yield workunits for various Dataplex resources."""
        # Iterate over all configured projects
        for project_id in self.config.project_ids:
            logger.info(f"Processing Dataplex resources for project: {project_id}")
            with self.report.new_stage(
                f"Processing entries from Universal Catalog for project {project_id}"
            ):
                try:
                    yield from auto_workunit(
                        self.entries_processor.process_project(project_id)
                    )
                except exceptions.GoogleAPICallError as exc:
                    self.report.warning(
                        title="Failed to process Dataplex entries",
                        message="Error while extracting entries from Universal Catalog.",
                        context=project_id,
                        exc=exc,
                    )
                    continue

        if self.config.include_lineage and self.lineage_extractor:
            with self.report.new_stage(
                "Extracting Dataplex lineage across configured projects"
            ):
                if len(self.entry_data) == 0:
                    logger.info(
                        "No entries found for lineage extraction across configured projects"
                    )
                    return

                logger.info(
                    "Extracting lineage for %s entries across %s configured projects",
                    len(self.entry_data),
                    len(self.config.project_ids),
                )
                lineage_project_location_pairs = (
                    self._resolve_lineage_project_location_pairs()
                )

                try:
                    yield from self.lineage_extractor.get_lineage_workunits(
                        self.entry_data,
                        active_lineage_project_location_pairs=lineage_project_location_pairs,
                    )
                except Exception as e:
                    self.report.warning(
                        title="Lineage extraction failed",
                        message="Failed to extract lineage across configured projects",
                        context="all-configured-projects",
                        exc=e,
                    )

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "DataplexSource":
        """Factory method to create DataplexSource instance."""
        config = DataplexConfig.model_validate(config_dict)
        return cls(ctx, config)



from typing import Dict, Optional

from pydantic import Field

from datahub.configuration.common import AllowDenyPattern
from datahub.configuration.kafka import KafkaConsumerConnectionConfig
from datahub.configuration.source_common import (
    DatasetSourceConfigMixin,
    LowerCaseDatasetUrnConfigMixin,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
)


class KafkaSourceConfig(
    StatefulIngestionConfigBase,
    DatasetSourceConfigMixin,
    LowerCaseDatasetUrnConfigMixin,
):
    connection: KafkaConsumerConnectionConfig = KafkaConsumerConnectionConfig()

    topic_patterns: AllowDenyPattern = AllowDenyPattern(allow=[".*"], deny=["^_.*"])
    domain: Dict[str, AllowDenyPattern] = Field(
        default={},
        description="A map of domain names to allow deny patterns. Domains can be urn-based (`urn:li:domain:13ae4d85-d955-49fc-8474-9004c663a810`) or bare (`13ae4d85-d955-49fc-8474-9004c663a810`).",
    )
    topic_subject_map: Dict[str, str] = Field(
        default={},
        description="Provides the mapping for the `key` and the `value` schemas of a topic to the corresponding schema registry subject name. Each entry of this map has the form `<topic_name>-key`:`<schema_registry_subject_name_for_key_schema>` and `<topic_name>-value`:`<schema_registry_subject_name_for_value_schema>` for the key and the value schemas associated with the topic, respectively. This parameter is mandatory when the [RecordNameStrategy](https://docs.confluent.io/platform/current/schema-registry/serdes-develop/index.html#how-the-naming-strategies-work) is used as the subject naming strategy in the kafka schema registry. NOTE: When provided, this overrides the default subject name resolution even when the `TopicNameStrategy` or the `TopicRecordNameStrategy` are used.",
    )
    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = None
    schema_registry_class: str = Field(
        default="datahub.ingestion.source.confluent_schema_registry.ConfluentSchemaRegistry",
        description="The fully qualified implementation class(custom) that implements the KafkaSchemaRegistryBase interface.",
    )
    schema_tags_field: str = Field(
        default="tags",
        description="The field name in the schema metadata that contains the tags to be added to the dataset.",
    )
    enable_meta_mapping: bool = Field(
        default=True,
        description="When enabled, applies the mappings that are defined through the meta_mapping directives.",
    )
    meta_mapping: Dict = Field(
        default={},
        description="mapping rules that will be executed against top-level schema properties. Refer to the section below on meta automated mappings.",
    )
    field_meta_mapping: Dict = Field(
        default={},
        description="mapping rules that will be executed against field-level schema properties. Refer to the section below on meta automated mappings.",
    )
    strip_user_ids_from_email: bool = Field(
        default=False,
        description="Whether or not to strip email id while adding owners using meta mappings.",
    )
    tag_prefix: str = Field(
        default="", description="Prefix added to tags during ingestion."
    )
    ignore_warnings_on_schema_type: bool = Field(
        default=False,
        description="Disables warnings reported for non-AVRO/Protobuf value or key schemas if set.",
    )
    disable_topic_record_naming_strategy: bool = Field(
        default=False,
        description="Disables the utilization of the TopicRecordNameStrategy for Schema Registry subjects. For more information, visit: https://docs.confluent.io/platform/current/schema-registry/serdes-develop/index.html#handling-differences-between-preregistered-and-client-derived-schemas:~:text=io.confluent.kafka.serializers.subject.TopicRecordNameStrategy",
    )
    ingest_schemas_as_entities: bool = Field(
        default=False,
        description="Enables ingesting schemas from schema registry as separate entities, in addition to the topics",
    )
    external_url_base: Optional[str] = Field(
        default=None,
        description="Base URL for external platform (e.g. Aiven) where topics can be viewed. The topic name will be appended to this base URL.",
    )
